/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <console.h>
#include <byteswap.h>
#include <curses.h>
#include <stdlib.h>
#include <limits.h>

#include <gpxe/command.h>
#include <gpxe/settings.h>
#include <gpxe/in.h>
#include <gpxe/editbox.h>
#include <gpxe/keys.h>
#include <gpxe/sf_settings.h>
#include <gpxe/shell.h>
#include <gpxe/process.h>
#include <gpxe/sf_widgets.h>
#include <gpxe/sf_widgets_run.h>

static void sf_dummy_void ( struct sf_widget *this __unused ) { };
static int sf_dummy_save ( struct sf_widget *this __unused,
			   struct sf_widget **error_out __unused ) {
	return 0;
}
static int sf_dummy_run ( struct sf_widget *this __unused, int key __unused ) {
	return 0;
}

/******************************************************************************/

static void sf_widget_init ( struct sf_widget *widget,
			     enum sf_widget_type type, 
			     struct sf_widget_operations *op,
			     struct settings *settings,
			     struct setting *setting,
			     WINDOW *wnd, const char *text,
			     unsigned col, unsigned row )
{
	widget->type = type;
	INIT_LIST_HEAD ( & ( widget->list ) );
	widget->op = op;
	widget->wnd = wnd;
	widget->carret = SF_CARRET_MARGIN;
	widget->visible = TRUE;
	widget->text = ( text != NULL ) ? strdup ( text ) : NULL;
	widget->settings = settings;
	widget->setting = setting;
	widget->col = col;
	widget->row = widget->ori = row;

	if ( setting && setting->type->storef )
		widget->selectable = TRUE;
	else
		widget->selectable = FALSE;
}

static void sf_widget_delete ( sf_widget *widget ) {
	widget->op->delete ( widget );

	free ( widget->text );
}

/**
 * Free widgets list
 *
 * @v list	List to free
 */
void sf_widgets_list_free ( sf_widget *list ) {
	sf_widget	*p;

	if ( list == NULL )
		return;

	while ( ! list_empty ( &list->list ) ) {
		p = next_widget ( list );
		list_del ( & ( p->list ) );
		sf_widget_delete ( p );
		free ( p );
	}

	sf_widget_delete ( list );
	free ( list );
}


void sf_widget_attach ( sf_widget *parent, sf_widget *child ) {
	list_join ( & ( parent->list ), & ( child->list ) );
}


int sf_widget_action ( sf_widget *widget ) {
	if ( widget->action == NULL )
		return 0;

	return widget->action ( widget );
}


/******************************************************************************/

static void sf_window_load ( sf_widget *this ) {
	sf_widget *w;

	for ( w = next_widget ( this ) ; w != this ; w = next_widget ( w ) )
		sf_widget_load ( w );
}

static int sf_window_save ( sf_widget *this, sf_widget **error_out ) {
	sf_widget	*w;
	int		rc;

	for ( w = next_widget ( this ) ; w != this ; w = next_widget ( w ) ) {
		if ( ( rc = sf_widget_save ( w, error_out ) ) < 0 )
			return rc;
	}

	return 0;
}

/**
 * Delete method for 'Window' widget
 *
 * @v this	The widget
 */
static void sf_window_delete ( sf_widget *this ) {
	free ( this->wnd );
}


static struct sf_widget_operations sf_window_operations = {
	.load		= sf_window_load,
	.save		= sf_window_save,
	.draw		= sf_window_draw,
	.scroll		= sf_window_scroll,
	.run		= sf_window_run,
	.delete		= sf_window_delete,
};


/**
 * Prepare sf_widget structure for 'Window' widget
 *
 * @v title		Window title
 * @v status		Status bar contents
 * @v row
 * @v column		Location
 * @v width
 * @v height		Window size
 * @v list		List of widgets to attach
 *
 * @ret			New window widget
 */
struct sf_widget *sf_window_init ( const char *text, unsigned int row,
				   unsigned int column, unsigned int width,
				   unsigned int height ) {
	WINDOW *wnd;
	sf_window *window;

	window = zalloc ( sizeof ( sf_window ) );
	if ( window == NULL )
		return NULL;

	wnd = newwin ( height, width, row, column );
	if ( wnd == NULL ) {
		free ( window );
		return NULL;
	}

	sf_widget_init ( &window->widget, SFW_WINDOW, &sf_window_operations,
			 NULL, NULL, wnd, text, column, row );
	window->widget.ori = 0;
	window->draw_border = 1;

	return &window->widget;
}

/******************************************************************************/

static struct sf_widget_operations sf_button_operations = {
	.load		= sf_dummy_void,
	.save		= sf_dummy_save,
	.draw		= sf_button_draw,
	.scroll		= NULL,
	.run		= sf_button_run,
	.delete		= sf_dummy_void,
};


/**
 * Init sf_widget structure for 'Button' widget
* 
 * @v label
 * @v wnd
 * @v row
 * @v column		Values for sf_button structure
 *
 * @ret			New button widget
 */
sf_widget *sf_button_init ( const char *label, WINDOW *wnd,
			    unsigned int row, unsigned int column ) {
        sf_button *button;

	button = zalloc ( sizeof ( sf_button ) );
	if ( button == NULL )
		return NULL;

	sf_widget_init ( &button->widget, SFW_BUTTON, &sf_button_operations,
			 NULL, NULL, wnd, label, column, row );
	button->widget.selectable = TRUE;

	return &button->widget;
}

/******************************************************************************/

/**
 * Load method for 'Button' widget
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
static void sf_editbox_load ( struct sf_widget *this )
{
	sf_editbox		*box = SF_EDITBOX ( this );
	struct edit_string	*string = &box->editbox.string;

	/* No guarantee that the underlying setting is really
	 * a string, so we can't use refetch_string */
	if ( this->setting == NULL ||
	     ( ( fetchf_setting ( this->settings, this->setting,
				  string->buf, string->len ) < 0 ) &&
	       ( fetchf_setting ( sf_default_settings, this->setting,
				  string->buf, string->len ) < 0 ) ) )
		string->buf[0] = '\0';

	string->cursor = strlen ( string->buf );
}


static int sf_editbox_save ( sf_widget *this, sf_widget **error_out )
{
	sf_editbox		*box = SF_EDITBOX ( this );
	struct edit_string	*string = &box->editbox.string;

	if ( this->setting == NULL || this->settings == NULL )
		return 0;

	*error_out = this;
	return storef_setting ( this->settings, this->setting,
				string->buf[0] == '\0' ? NULL : string->buf );
}


/**
 * Delete method for 'Editbox' widget
 *
 * @v this	The widget
 */
static void sf_editbox_delete ( sf_widget *this ) {
	sf_editbox *sf_box = SF_EDITBOX ( this );

	free ( sf_box->editbox.string.buf );
	sf_box->editbox.string.buf = NULL;
}


static struct sf_widget_operations sf_editbox_operations = {
	.load		= sf_editbox_load,
	.save		= sf_editbox_save,
	.draw		= sf_editbox_draw,
	.scroll		= sf_editbox_scroll,
	.run		= sf_editbox_run,
	.delete		= sf_editbox_delete,
};


/**
 * Init sf_widget structure for 'Editbox' widget
 *
 * @v wnd
 * @v row
 * @v column		Editbox location
 * @v len		Editbox buffer length
 * @v desc		Description
 * @v desc_col		Column where to print description
 * @v settings
 * @v setting		Core settings structures
 * @v width		Editbox width
 *
 * @ret		        Editbox widget
 */
sf_widget *sf_editbox_init ( WINDOW *wnd, unsigned int row,
			     unsigned int column, size_t len,
			     const char *desc, struct settings *settings,
			     struct setting *setting, unsigned int width ) {
	sf_editbox	*sf_box;
	char		*buf;

	sf_box = zalloc ( sizeof ( sf_editbox ) );
	if ( sf_box == NULL )
		return NULL;

	buf = malloc ( len );
	if ( buf == NULL ) {
		free ( sf_box );
		return NULL;
	}

	sf_widget_init ( &sf_box->widget, SFW_EDITBOX,  &sf_editbox_operations,
			 settings, setting, wnd, desc, column, row );
	sf_box->shadow = '\0';

	buf[0] = '\0';
	if ( desc != NULL )
		column += strlen ( desc );
	init_editbox ( &sf_box->editbox, buf, len, wnd, row, column, width, 0 );

	return &sf_box->widget;
}

/******************************************************************************/

/**
 * Load method for 'Checkbox' widget
 *
 * @v this	The widget
 */
static void sf_checkbox_load ( struct sf_widget *this ) {
	sf_checkbox	*box = SF_CHECKBOX ( this );
	long		value = sf_sett_get_intz ( this->settings,
						   this->setting );

	box->checked = ( value ? TRUE : FALSE );
}


static int sf_checkbox_save ( sf_widget *this, sf_widget **error_out ) {
	sf_checkbox	*box = SF_CHECKBOX ( this );
	long 		value = box->checked ? 1 : 0;

	if ( this->setting == NULL || this->settings == NULL )
		return 0;

	*error_out = this;
	return store_setting ( this->settings, this->setting, &value, 1 );
}


static struct sf_widget_operations sf_checkbox_operations = {
	.load		= sf_checkbox_load,
	.save		= sf_checkbox_save,
	.draw		= sf_checkbox_draw,
	.scroll		= NULL,
	.run		= sf_checkbox_run,
	.delete		= sf_dummy_void,
};


/**
 * Prepare sf_widget structure for 'Checkbox' widget
 *
 * @v wnd
 * @v row
 * @v col		Checkbox location
 * @v desc		Description
 * @v data		Callback data for action() method
 * @v settings
 * @v setting		Core settings structures
 *
 * @ret			new checkbox widget
 */
sf_widget *sf_checkbox_init ( WINDOW *wnd, int row, int column,
			      const char *desc, struct settings *settings,
			      struct setting *setting ) {
	sf_checkbox *box;

	box = zalloc ( sizeof ( sf_checkbox ) );
	if ( box == NULL )
		return NULL;

	sf_widget_init ( &box->widget, SFW_CHECKBOX, &sf_checkbox_operations,
			 settings, setting, wnd, desc, column, row );

	return &box->widget;
}

/******************************************************************************/

/**
 * Load method for 'Combo' widget
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
static void sf_combo_load ( struct sf_widget *this ) {
	sf_combo *combo = SF_COMBO ( this );
	long key;
	int index;

	if ( this->settings == NULL )
		goto fail;

	key = sf_sett_get_intz ( this->settings, this->setting );

	for ( index = 0 ; index < combo->varn; ++index ) {
		if ( key == combo->vars[index].key ) {
			combo->index = index;
			return;
		}
	}

fail:
	combo->index = 0;
}


static int sf_combo_save ( sf_widget *this, sf_widget **error_out ) {
	sf_combo	*combo = SF_COMBO ( this );
	long		index = combo->index;
	char            buf[20];

	if ( this->setting == NULL || this->settings == NULL )
		return 0;

	/* So that we can store the setting in the correct width, we
	 * serialise the setting to a string, then storef the value */
	snprintf ( buf, sizeof ( buf ), "%ld", combo->vars[index].key );

	*error_out = this;
	return storef_setting ( this->settings, this->setting, buf );
}

static struct sf_widget_operations sf_combo_operations = {
	.load		= sf_combo_load,
	.save		= sf_combo_save,
	.draw		= sf_combo_draw,
	.scroll		= NULL,
	.run		= sf_combo_run,
	.delete		= sf_dummy_void,
};


/**
 * Initialize an sf_combo structure for a 'Combo' widget
 * @v wnd
 * @v row
 * @v column		Combo location
 * @v desc		Description
 * @v settings
 * @v setting		Core settings structures
 * @v vars		Array of choice variants
 * @v varn		Choice variants quantity
 *
 * @ret			New Combo widget
 */
sf_widget *sf_combo_init ( WINDOW *wnd, int row, int column, const char *desc,
			   struct settings *settings, struct setting *setting,
			   const sf_option *vars, int varn ) {
	sf_combo *combo;

	combo = zalloc ( sizeof ( sf_combo ) );
	if ( combo == NULL )
		return NULL;

	sf_widget_init ( &combo->widget, SFW_COMBO, &sf_combo_operations,
			 settings, setting, wnd, desc, column, row );
	combo->vars = vars;
	combo->varn = varn;

	return &combo->widget;
}
	
/******************************************************************************/

static struct sf_widget_operations sf_text_operations = {
	.load		= sf_dummy_void,
	.save		= sf_dummy_save,
	.draw		= sf_text_draw,
	.scroll		= NULL,
	.run		= sf_dummy_run,
	.delete		= sf_dummy_void,
};


/* Initialize an sf_widget structure for a 'static text' widget
 * @v wnd
 * @v row
 * @v column		Widget location
 * @v text		Text contents
 *
 * @ret			New text widget
 */
struct sf_widget *sf_text_init ( WINDOW *wnd, int row,
				 int column, const char *text ) {
	sf_text *box;

	box = zalloc ( sizeof ( sf_text ) );
	if ( box == NULL )
		return NULL;

	sf_widget_init ( &box->widget, SFW_TEXT, &sf_text_operations,
			 NULL, NULL, wnd, text, column, row );
	return &box->widget;
}

/******************************************************************************/

/**
 * Delete method for 'Info' widget
 *
 * @v this	The widget
 */
static void sf_info_delete ( sf_widget *this ) {
	sf_info *info = SF_INFO ( this );

	free ( info->value );
}


static struct sf_widget_operations sf_info_operations = {
	.load		= sf_dummy_void,
	.save		= sf_dummy_save,
	.draw		= sf_info_draw,
	.scroll		= NULL,
	.run		= sf_dummy_run,
	.delete		= sf_info_delete,
};


/* Initialize an sf_widget structure for a 'static text' widget
 * @v wnd
 * @v row
 * @v column		Widget location
 * @v text		Text contents
 *
 * @ret			New text widget
 */
struct sf_widget *sf_info_init ( WINDOW *wnd, unsigned int row,
				 unsigned int col, const char *text,
				 const char *value ) {
	sf_info *info;

	info = zalloc ( sizeof ( sf_info ) );
	if ( info == NULL )
		return NULL;

	sf_widget_init ( &info->widget, SFW_INFO, &sf_info_operations,
			 NULL, NULL, wnd, text, col, row );
	info->value = ( value != NULL ) ? strdup ( value ) : NULL;
	info->last_length = 0;

	return &info->widget;
}

/******************************************************************************/


static void sf_list_load ( sf_widget *this ) {
	sf_list *list = SF_LIST ( this );

	sf_widget_load ( list->child );
}


static int sf_list_save ( sf_widget *this, sf_widget **error_out ) {
	sf_list *list = SF_LIST ( this );

	return sf_widget_save ( list->child, error_out );
}


static void sf_list_delete ( sf_widget *this ) {
	sf_list *list = SF_LIST ( this );

	sf_widgets_list_free ( list->child );
}


static struct sf_widget_operations sf_list_operations = {
	.load		= sf_list_load,
	.save		= sf_list_save,
	.draw		= sf_list_draw,
	.scroll		= sf_list_scroll,
	.run		= sf_list_run,
	.delete		= sf_list_delete,
};


/* Initialize an sf_widget structure for a 'list' widget */
struct sf_widget *sf_list_init ( WINDOW *wnd, unsigned int row,
				 unsigned int column, unsigned int width,
				 unsigned int height ) {
	sf_list	 	*list;
	sf_window	*window;

	list = zalloc ( sizeof ( sf_info ) );
	if ( list == NULL )
		return NULL;

	list->child = sf_window_init ( NULL, row + wnd->ori_y,
				       column + wnd->ori_x,
				       width, height );
	if ( list->child == NULL ) {
		free ( list );
		return NULL;
	}
	window = SF_WINDOW ( list->child );
	window->draw_border = 0;
	window->scrollable = 1;

	sf_widget_init ( &list->widget, SFW_LIST, &sf_list_operations,
			 NULL, NULL, wnd, NULL, column, row );
	list->widget.selectable = TRUE;

	return &list->widget;
}

/******************************************************************************/

static void sf_row_load ( struct sf_widget *this )
{
	int scroll = this->ori - this->row;
	uint8_t value;

	value = sf_sett_get_intz ( this->settings, this->setting );
	/* A row setting is 1..255, but cope with an erroneous row = 0 */
	if ( value == 0 )
		value = 1;
	this->row = value - 1;
	this->ori = this->row + scroll;
}

static int sf_row_save ( sf_widget *this, sf_widget **error_out )
{
	uint8_t value = this->row + 1;

	if ( this->setting == NULL || this->settings == NULL )
		return 0;

	*error_out = this;
	return store_setting ( this->settings, this->setting, &value, 1 );
}


static struct sf_widget_operations sf_row_operations = {
	.load		= sf_row_load,
	.save		= sf_row_save,
	.draw		= sf_row_draw,
	.scroll		= NULL,
	.run		= sf_row_run,
	.delete		= sf_dummy_void,
};


/**
 * Prepare sf_widget structure for 'Row' widget
 *
 * @v wnd
 * @v col		Row column
 * @v text		Text identifying row
 * @v settings
 * @v setting		Core settings structures
 *
 * @ret			new checkbox widget
 */
sf_widget *sf_row_init ( WINDOW *wnd, unsigned int column, const char *text,
			 struct settings *settings, struct setting *setting )
{
	sf_row *row;

	row = zalloc ( sizeof ( *row ) );
	if ( row == NULL )
		return NULL;

	sf_widget_init ( &row->widget, SFW_ROW, &sf_row_operations,
			 settings, setting, wnd, text, column, 0 );

	return &row->widget;
}
