/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <console.h>
#include <byteswap.h>
#include <curses.h>
#include <stdlib.h>
#include <limits.h>

#include <gpxe/command.h>
#include <gpxe/settings.h>
#include <gpxe/in.h>
#include <gpxe/editbox.h>
#include <gpxe/keys.h>
#include <gpxe/sf_settings.h>
#include <gpxe/shell.h>
#include <gpxe/process.h>
#include <gpxe/sf_widgets.h>
#include <gpxe/sf_main.h>

#define ARRAY_SIZE( _x ) ( sizeof ( _x ) / sizeof ( _x[0] ) )

#define WINDOW_TITLE							\
	" " PRODUCT_NAME " "
#define WINDOW_STATUS							\
	"Up,Down Arrow to select | SPACE,+,- to change | ESC to exit | F1=help"

static int draw_disabled = 0;

static void
draw_carret ( sf_widget *widget)
{
	attr_t attr = ATTR_TEXT | COLOR_PAIR ( CPAIR_TEXT );

	if ( !widget->selectable )
		return;

	if ( widget->carret == SF_CARRET_MARGIN ) {
		mvwaddch ( widget->wnd, widget->ori, 2,
			   attr | ( widget->selected ? '>' : ' ' ) );
	} else if ( widget->carret == SF_CARRET_WINDOW ) {
		mvwaddch ( widget->wnd, widget->ori, widget->col - 2,
			   attr | ( widget->selected ? '>' : ' ' ) );
		mvwaddch ( widget->wnd, widget->ori,
			   widget->col + strlen ( widget->text ) + 1,
			   attr | ( widget->selected ? '<' : ' ' ) );
	}
}

static inline __always_inline void
set_attr ( WINDOW *wnd, unsigned int attr, unsigned int cpair )
{
	wnd->attrs = ( attr | COLOR_PAIR ( cpair ) );
}

static void
set_attr_selectable ( sf_widget *widget )
{
	if ( widget->selectable && widget->selected )
		set_attr ( widget->wnd, ATTR_CURRENT, CPAIR_CURRENT );
	else if ( widget->selectable )
		set_attr ( widget->wnd, ATTR_TEXT, CPAIR_TEXT );
	else if ( widget->visible )
		set_attr( widget->wnd, ATTR_INFO, CPAIR_INFO );
	else
		set_attr ( widget->wnd, ATTR_DISABLED, CPAIR_DISABLED );
}

static void widget_waddstr ( sf_widget *this, bool visible, const char *text )
{
	attr_t		attrs = this->wnd->attrs;
	const char	*p;

	for ( p = text; *p != '\0'; ++p )
		waddch ( this->wnd, attrs | ( visible ? *p : ' ' ) );
}

static void
widget_mvwaddstr ( sf_widget *this, const char *text )
{
	wmove ( this->wnd, this->ori, this->col );
	widget_waddstr ( this, this->visible, text );
}

static int
window_mvwaddstr(WINDOW *win, int y, int x, const char *text)
{
	if ( x < 0 )
		x = ( win->width - strlen ( text ) ) / 2;

	return mvwaddstr ( win, y, x, text );
}

/* 'Widget' implementation */

static void sf_widget_scroll ( sf_widget *this ) {
	if ( this->op->scroll != NULL )
		this->op->scroll ( this );
}

static int sf_widget_default_run ( int key ) {
	switch ( key ) {
	case TAB:
		return RC_TAB;
	case KEY_UP:
	case CTRL_K:
		return RC_GOUP;
	case KEY_DOWN:
		return RC_GODOWN;
	case KEY_LEFT:
	case CTRL_H:
		return RC_GOLEFT;
	case KEY_RIGHT:
	case CTRL_L:
		return RC_GORIGHT;
	default:
		return 0;
	}		
}

/* A list of all the currently running windows. Allows sf_window_run()
 * to figure out which windows need to be redrawn as windows are unhidden
 */
static LIST_HEAD ( window_list );


/** Redraw enough of the window stack so that window is totally covered */
static void redraw_stack ( sf_window *window ) {
	struct list_head *pos;

	/* Set the cursor to the top most window */
	curs_set ( list_entry ( window_list.prev, struct sf_window, list )->cursor );

	if ( window == NULL )
		pos = window_list.next;
	else {
		WINDOW *wnd = window->widget.wnd;

		unsigned parts = 0;

		for ( pos = window_list.next; pos != &window_list;
		      pos = pos->next ) {
			WINDOW *other = list_entry ( pos, struct sf_window,
						     list )->widget.wnd;

			if ( other->ori_x <= wnd->ori_x )
				parts |= 1;
			if ( other->ori_y <= wnd->ori_y )
				parts |= 2;
			if ( other->ori_x + other->width >= wnd->ori_x + wnd->width )
				parts |= 4;
			if ( other->ori_y + other->height >= wnd->ori_y + wnd->height )
				parts |= 8;
			if ( parts == 0xf )
				break;
		}
	}

	/* Redraw each window */
	for ( ; pos != &window_list ; pos = pos->next )
		sf_widget_draw ( &list_entry ( pos, struct sf_window, list )->widget );
}

/**
 * A simple wrapper for mvwaddstr with auto truncation and centering.
 *
 * @v window	The 'window' widget
 * @v y     	Y-coordinate
 * @v x     	X-coordinate: -1, means centering.
 * @v fmt   	format string.
 *
 * @ret     	Status code
 */
/*
 * 'Window' widget implementation
 */
static void sf_window_draw_border ( sf_widget *this ) {
	sf_window	*window = SF_WINDOW ( this );
	WINDOW		*wnd = window->widget.wnd;
	chtype		attrs;
	int		title_row;

	set_attr ( wnd, ATTR_TEXT, CPAIR_TEXT );
	attrs = wnd->attrs;

	/* Draw border */
	if ( window->draw_border ) {
		wborder ( wnd, ACS_DVLINE | attrs, ACS_DVLINE | attrs,
			  ACS_DHLINE | attrs, ACS_DHLINE | attrs,
			  ACS_ULCORNER | attrs, ACS_URCORNER | attrs,
			  ACS_LLCORNER | attrs, ACS_LRCORNER | attrs );
	}

	/* Draw window title */
	if ( wnd->ori_x == 0 ) {
		window_mvwaddstr ( wnd, 0, -1, WINDOW_TITLE );
		window_mvwaddstr ( wnd, wnd->height - 1, -1, WINDOW_STATUS );

		title_row = 1;
	} else
		title_row = 0;

	/* Draw window text */
	if ( window->widget.text != NULL ) {
		window_mvwaddstr ( wnd, title_row, -1, this->text );
		this->row = title_row;
	}

	/* Draw scrollbar */
	if ( window->scrollable ) {
		int		min, max, region;
		sf_widget	*w;

		/* Count the total number of content rows */
		min = INT_MAX;
		max = INT_MIN;
		for ( w = next_widget ( this ); w != this; w = next_widget ( w ) ) {
			if ( w->row < min && w->visible )
				min = w->row;
			if ( w->row > max && w->visible )
				max = w->row;
		}

		/* A scrollbar is only required if the total number
		 * of rows is greater than the visible area
		 */
		region = ( max + 1 - min - wnd->height );
		if ( region > 0 ) {
			int position;

			/* Draw the entire scroll bar */
			wmove ( wnd, 0, wnd->width - 1 );
			wvline ( wnd, 177 | attrs, wnd->height );

			/* Draw the scroll position */
			position = -this->ori * ( wnd->height - 1 ) / region;
			mvwaddch ( wnd, position, wnd->width - 1,
				   wnd->attrs | '\333' );
		}
	}
}

static int sf_dialog_ok ( sf_widget *this __unused )
{
	return RC_QUIT;
}

static int sf_dialog_cancel ( sf_widget *this __unused )
{
	return RC_COMPLETE;
}

/**
 * Show an informational dialog
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
int sf_dialog ( sf_widget **dialog, const char *title,
		enum sf_dialog_type dialog_type, int width, int height,
		const char **lines, unsigned int line_count )
{
	WINDOW *parent;
	sf_widget *wnd, *list = NULL;
	const char *caption;
	unsigned int line;
	int rc, pos;

	if ( height == -1 ) {
		height = line_count + 4;
		if ( dialog_type != SFW_DIALOG_MSG )
			height += 3;
	}
	if ( width == -1 ) {
		width = 34;
		for ( line = 0; line < line_count; line++ ) {
			if ( lines[line] != NULL
			     && strlen ( lines[line] ) > ( unsigned ) width )
				width = strlen ( lines[line] );
		}
		width += 6;
	}

	/* Draw in top window */
	parent = list_entry ( window_list.next,
			      struct sf_window, list )->widget.wnd;
	wnd = sf_window_init ( title,
			       parent->ori_y + ( int )( parent->height
							- height ) / 2,
			       parent->ori_x + ( int )( parent->width
							- width ) / 2,
			       width, height );
	if ( wnd == NULL )
		return -ENOMEM;
	if ( dialog != NULL )
		*dialog = wnd;

	for ( line = 0; line < line_count; line++ ) {
		if ( lines[line] == NULL )
			continue;

		list = sf_text_init ( wnd->wnd, line + 2, 3, lines[line] );
		if ( list == NULL ) {
			rc = -ENOMEM;
			goto out;
		}
		sf_widget_attach ( wnd, list );
	}

	/* Draw the OK button */
	caption = NULL;
	pos = 0;
	switch ( dialog_type ) {
	case SFW_DIALOG_MSG:
		break;

	case SFW_DIALOG_OKCANCEL:
		caption = "  OK  ";
		pos = ( width / 2 ) - 4;

		/* fall-thru */
	case SFW_DIALOG_EDITCANCEL:
		if ( caption == NULL )
			caption = " Edit ";
		pos = ( width / 2 ) - 11;

		/* fall-thru */
	case SFW_DIALOG_OK:
		if ( caption == NULL ) {
			caption = "  OK  ";
			pos = ( width / 2 ) - 4;
		}

		list = sf_button_init ( caption, wnd->wnd, line + 3, pos );
		if ( list == NULL ) {
			rc = -ENOMEM;
			goto out;
		}
		list->action = sf_dialog_ok;
		list->carret = SF_CARRET_NONE;
		sf_widget_attach ( wnd, list );
	}

	/* Draw the cancel button */
	if ( dialog_type == SFW_DIALOG_OKCANCEL ||
	     dialog_type == SFW_DIALOG_EDITCANCEL ) {
		/* Use windowed carret for both buttons */
		list->carret = SF_CARRET_WINDOW;

		list = sf_button_init ( "Cancel", wnd->wnd, line + 3,
					( width / 2 ) + 3 );
		if ( list == NULL ) {
			rc = -ENOMEM;
			goto out;
		}
		list->action = sf_dialog_cancel;
		list->carret = SF_CARRET_WINDOW;
		sf_widget_attach ( wnd, list );
	}

	rc = sf_widget_run ( wnd, 0 );

	if ( rc == RC_QUIT ) {
		/* sf_widget_run() assumed that we were going to
		 * quit. Redraw the window stack */
		redraw_stack ( SF_WINDOW ( wnd ) );
		rc = 0;
	} else if ( dialog_type == SFW_DIALOG_OK )
		rc = 0;
	else
		rc = RC_REVERT;

out:
	if ( dialog != NULL )
		*dialog = NULL;
	sf_widgets_list_free ( wnd );
	return rc;
}

void sf_window_draw ( sf_widget *this ) {
	sf_widget *w, *selected = NULL;

	set_attr ( this->wnd, ATTR_TEXT, CPAIR_TEXT );
	werase ( this->wnd );

	/* Checkbox/Combo actions affect visibility of other controls,
	 * so run their actions upfront */
	draw_disabled = 1;
	for ( w = next_widget ( this ) ; w != this; w = next_widget ( w ) ) {
		if ( w->type == SFW_CHECKBOX || w->type == SFW_COMBO )
			sf_widget_action ( w );
	}
	draw_disabled = 0;

	for ( w = next_widget ( this ) ; w != this; w = next_widget ( w ) ) {
		if ( w->visible && w->ori >= 0 &&
		     w->ori <  ( int ) w->wnd->height ) {
			/* Draw the selected element last so the editbox cursor
			 * is positioned correctly (rather than double drawing
			 * the selected entry).
			 */
			if ( w->selected ) {
				assert ( selected == NULL );
				selected = w;
			} else
				sf_widget_draw ( w );
		}
	}

	sf_window_draw_border ( this );

	if ( selected != NULL )
		sf_widget_draw ( selected );
}

void sf_window_scroll ( sf_widget *this ) {
	sf_widget *w;

	for ( w = next_widget ( this ); w != this; w = next_widget ( w ) ) {
		/* Window overloads 'ori', and uses it as a scroll */
		w->ori = w->row + this->ori;
		sf_widget_scroll ( w );
	}

	/* Adjust the scroll bar */
	sf_window_draw_border ( this );
}

static void sf_widget_enter ( sf_widget *widget ) {
	widget->selected = TRUE;
	if ( widget->enter != NULL )
		widget->enter ( widget );
}

static void sf_widget_exit ( sf_widget *widget ) {
	widget->selected = FALSE;
	if ( widget->exit != NULL )
		widget->exit ( widget );
}

static void sf_widget_help ( sf_widget *widget ) {
	const int	width = 64, span = 58;
	int		over_length, end_of_string, new_line, row_count;
	char		*buf, *line_start, *last_space, *p;
	const char	*rows[20];
	char		title[40];

	if ( widget->help == NULL )
		return;

	if ( widget->setting != NULL )
		snprintf ( title, sizeof ( title ) - 1, " Help: %s ",
			   widget->setting->description );
	else
		strcpy ( title, " Help " );

	/* Paginate the help string into a list of rows rows */
	row_count = 0;
	buf = line_start = last_space = p = strdup ( widget->help );
	for ( p = buf; ; p++ ) {
		if ( *p == ' ' ) {
			last_space = p;
			continue;
		}

		over_length = ( p - line_start > span );
		end_of_string = ( *p == '\0' );
		new_line = ( *p == '\n' );

		if ( new_line )
			*p = '\0';
		else if ( over_length )
			*last_space = '\0';

		if ( over_length || end_of_string || new_line ) {
			rows[row_count++] = line_start;
			if ( new_line )
				line_start = last_space = p + 1;
			else if ( over_length )
				line_start = last_space = last_space + 1;
			else if ( end_of_string )
				break;
		}
	}

	(void) sf_dialog ( NULL, title, SFW_DIALOG_OK,
			   width, row_count + 6, rows, row_count );
	free ( buf );
}

/* Adjust the viewport onto the scrolled region if necessary. Returns
 * true if the window was scrolled */
static int sf_window_viewport ( sf_widget *this, sf_window *window )
{
	WINDOW *wnd = window->widget.wnd;

	if ( !window->scrollable )
		return 0;

	if ( this->ori < 0 ) {
		/* Scroll up, which means move everything down */
		window->widget.ori += - this->ori;
		sf_window_scroll ( &window->widget );
		return 1;

	} else if ( this->ori > ( signed ) wnd->height ) {
		/* Scroll down, which means move everything up */
		window->widget.ori += wnd->height - 2 - this->ori;
		sf_window_scroll ( &window->widget );
		return 1;
	}

	return 0;
}

static int sf_window_get_key ( void ) {	
	while ( 1 ) {
		if ( iskey() )
			return getkey ( 1 );

		step();
	}
}

/**
 * Run method for 'Window' widget
 *
 * @v this	The widget
 * @v key	Key: if non-zero, ask for confirmation before exit on ESC
 *
 * @ret		Status code
 */
int sf_window_run ( sf_widget *this, int key __unused ) {
	int		rc = 0, k;
	sf_widget	*list;
	sf_widget	*w, *new, *next;
	sf_window	*window = SF_WINDOW ( this );
	WINDOW		*wnd = window->widget.wnd;
	char		*prev_value = NULL;
	char		title[40];

	static const char *invalid_msg[] = {
		"Invalid value. You can either continue ",
		"editing this setting, or cancel your ",
		"changes and restore the original value.",
	};
	sf_widget_enter ( this );

	/* Find the first editable control select it, and scroll to it */
	for ( list = next_widget ( this );
	      ( !list->selectable && list->type != SFW_WINDOW );
	      list = next_widget ( list ) )
		;

	if ( list->type != SFW_WINDOW ) {
		sf_window_viewport ( list, window );
		sf_widget_enter ( list );
	}

	window->cursor = ( list->type == SFW_EDITBOX );
	curs_set ( window->cursor );
	sf_widget_draw ( this );
	list_add_tail ( &window->list, &window_list );

	/* If there aren't any visible controls, then just wait for ESC */
	if ( list->type == SFW_WINDOW ) {
		do {
			k = sf_window_get_key ( );
		} while ( k != ESC );

		rc = 0;
		goto finished;
	}

	/* A list control grabs and holds focus, mostly because nobody has
	 * suggested a good way of controlling enter/exit
	 */
	if ( list->type == SFW_LIST ) {
		rc = sf_widget_run ( list, 0 );
		goto finished;
	}

	if ( list->type == SFW_EDITBOX )
		prev_value = strdup ( sf_editbox_get ( SF_EDITBOX ( list ) ) );

wait_for_input:
	k = sf_window_get_key ( );

#ifndef SHELL_DISABLE
	if ( k == CTRL_S ) {
		/* Run shell [which nicely takes care
		 * of recursion for us] */
		curs_set ( 1 );
		endwin ( );
		shell ( );
		initscr ( );
		curs_set ( window->cursor );
		redraw_stack ( NULL );
		goto wait_for_input;
	}
#endif
#if defined(GDBSERIAL) || defined(GDBUDP)
	/* Provide a way of starting GDB stub when there's no shell */
	if ( k == CTRL_G ) {
		sf_widget_enter(list);
		curs_set ( 1 );
		endwin ( );
		sf_gdbstub ( );
		initscr ( );
		curs_set ( window->cursor );
		redraw_stack ( NULL );
		goto wait_for_input;
	}
#endif
	if ( k == KEY_F1 && list->help != NULL ) {
		sf_widget_help ( list );
		sf_widget_draw ( list );
		goto wait_for_input;
	}

	rc = sf_widget_run ( list, k );

	switch ( rc ) {
	case RC_INVAL:
		if ( list->setting != NULL )
			snprintf ( title, sizeof ( title ) - 1,
				   " Error: %s ", list->setting->description );
		else
			strcpy ( title, "Error" );
		rc = sf_dialog (
			NULL, title, SFW_DIALOG_EDITCANCEL, -1, -1,
			( const char ** )&invalid_msg,
			ARRAY_SIZE ( invalid_msg ) );

		if ( rc == RC_REVERT && list->type == SFW_EDITBOX ) {
			struct sf_editbox *box = SF_EDITBOX ( list );
			if ( strcmp ( box->editbox.string.buf,
				      prev_value ) != 0 ) {
				sf_editbox_set ( box, prev_value );
				sf_widget_draw ( list );
			}

			/* TODO: Support for RC_REVERT elsewhere */
		}

		/* User can't leave this control */
		goto wait_for_input;

	case RC_GOUP:
		for ( w = prev_widget ( list );
		      ( w != list &&
			( !w->selectable || w->ori >= list->ori ||
			  ( !window->scrollable && w->ori < 0 ) ) );
		      w = prev_widget ( w ) )
			;
		for ( next = prev_widget ( w );
		      ( w != list && next->ori == w->ori &&
			w->col > list->col ) ;
		      w = next, next = prev_widget ( w ) )
			;
		if ( w != list ) {
			new = w;
			goto new_control;
		}
		goto wait_for_input;

	case RC_GODOWN:
		for ( w = next_widget ( list );
		      ( w != list &&
			( !w->selectable || w->ori <= list->ori ||
			  ( !window->scrollable &&
			    w->ori >= ( signed ) wnd->height ) ) );
		      w = next_widget ( w ) )
			;
		for ( next = next_widget ( w );
		      ( w != list && next->selectable &&
			next->ori == w->ori && w->col < list->col ) ;
		      w = next, next = next_widget ( w ) )
			;
		if ( w != list ) {
			new = w;
			goto new_control;
		}
		goto wait_for_input;

	case RC_GOLEFT:
		for ( w = prev_widget ( list ); !w->selectable;
		      w = prev_widget ( w ) )
			;
		if ( w->ori == list->ori && w->col < list->col ) {
			new = w;
			goto new_control;
		}
		goto wait_for_input;

	case RC_GORIGHT:
		for ( w = next_widget ( list ); !w->selectable;
		      w = next_widget ( w ) )
			;
		if ( w->ori == list->ori && w->col > list->col ) {
			new = w;
			goto new_control;
		}
		goto wait_for_input;

	case RC_TAB:
		for ( w = next_widget ( list ); !w->selectable;
		      w = next_widget ( w ) )
			;
		if ( w != list ) {
			new = w;
			goto new_control;
		}
		goto wait_for_input;

	case RC_COMPLETE:
		rc = 0;
		break;

	case RC_QUIT:
		rc = RC_QUIT;
		break;
	default:
		if ( k != ESC )
			goto wait_for_input;
		rc = 0;
		break;
	}

	/* Predecate leaving the window on the window action */
	if ( this->action != NULL &&
	     this->action ( this ) != RC_QUIT )
		goto wait_for_input;

	if ( list->type == SFW_EDITBOX )
		free ( prev_value );

	sf_widget_exit ( list );

finished:
	/* This window had *better* be at the top of the stack! */
	assert ( window->list.next == &window_list );
	list_del ( &window->list );
	sf_widget_exit ( this );
	if ( this->transient )
		this->transient = FALSE;
	else if ( rc != RC_QUIT )
		/* There's no point redrawing the window stack
		 * if we're about to quit all the windows */
		redraw_stack ( window );

	return rc;

new_control:
	assert ( new != NULL );

	sf_widget_exit ( list );
	sf_widget_enter ( new );
	window->cursor = ( new->type == SFW_EDITBOX );
	curs_set ( window->cursor );

	if ( list->type == SFW_EDITBOX )
		free ( prev_value );
	if ( new->type == SFW_EDITBOX )
		prev_value = strdup ( sf_editbox_get ( SF_EDITBOX ( new ) ) );

	/* Scroll the window to the new control if need be */
	if ( sf_window_viewport ( new, window ) )
		sf_widget_draw ( this );
	else {
		sf_widget_draw ( list );
		sf_widget_draw ( new );
	}

	list = new;
	goto wait_for_input;
}



/**
 * Draw method for 'Button' widget
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
void sf_button_draw ( struct sf_widget *this ) {
	char *text = strdup ( this->text );
	unsigned len;

	/* If the final two characters of the button are "->", then
	 * draw these *only* if the button is selectable. */
	len = strlen ( this->text );
	if ( !this->selectable && len > 2 &&
	     text[len - 1] == '>' && text[len - 2] == '-' )
		text [ len - 2 ] = text [ len - 1 ] = ' ';

	draw_carret ( this );
	set_attr_selectable ( this );
	widget_mvwaddstr ( this, text );
	free ( text );
}

/**
 * Run method for 'Window' widget
 *
 * @v this	The widget
 * @v key	Key pressed to process
 *
 * @ret		Status code
 */
int sf_button_run ( struct sf_widget *this, int key ) {
	switch( key ) {
	case CR:
	case LF:
	case ' ':
		return sf_widget_action ( this );
	default:
		return sf_widget_default_run ( key );
	}
}

/******************************************************************************/

/**
 * Draw method for 'Button' widget
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
void sf_editbox_draw ( struct sf_widget *this ) {
	sf_editbox	*box = SF_EDITBOX ( this );
	struct edit_box	*ebox = &box->editbox;
	char		*tmp_buf = NULL;

	if ( draw_disabled )
		return;

	draw_carret ( this );
	set_attr_selectable ( this );
	widget_mvwaddstr ( this, this->text );

	if ( this->selectable ) {
		if ( box->shadow != '\0' ) {
			char *p;
			tmp_buf = strdup ( ebox->string.buf );
			for ( p = tmp_buf; *p != '\0'; p++ )
				*p = box->shadow;
			/* Save original buffer */
			p = ebox->string.buf;
			/* Replace it with shadowed string */
			ebox->string.buf = tmp_buf;
			draw_editbox ( ebox );
			ebox->string.buf = p;
			free ( tmp_buf );
		}
		else
			draw_editbox ( ebox );
	} else if (this->setting == &password_setting) {
		WINDOW *window = ebox->win;
		unsigned int i;

		/* Paint disabled editbox */
		wmove ( window, ebox->row, ebox->col );
		for (i = 0; i < ebox->width; i++)
			waddch ( window, window->attrs | ' ' );
	} else {
		mvwprintw ( ebox->win, ebox->row, ebox->col, "%s",
			    ebox->string.buf );
	}
}

void sf_editbox_scroll ( struct sf_widget *this ) {
	sf_editbox *box = SF_EDITBOX ( this );

	box->editbox.row = this->ori;
}

/**
 * Run method for 'Editbox' widget
 *
 * @v this	The widget
 * @v key	Key pressed to process
 *
 * @ret		Status code
 */
int sf_editbox_run ( struct sf_widget *this, int key ) {
	int		rc;
	sf_editbox	*box = SF_EDITBOX ( this );

	( void ) edit_editbox ( &box->editbox, key );

	sf_editbox_draw ( this );
	switch ( key ) {
	case ESC:
	case CR:
	case LF:
	case KEY_UP:
	case CTRL_K:
	case KEY_DOWN:
	case TAB:
		if ( ( rc = sf_widget_action ( this ) ) != 0 )
			return rc;

		return sf_widget_default_run ( key );

	default:
		return 0;
	}
}


void sf_editbox_set ( sf_editbox *box, char *value ) {
	strncpy ( box->editbox.string.buf, value, box->editbox.string.len );
	box->editbox.string.cursor = strlen ( value );
}

void sf_editbox_set_cursor ( sf_editbox *box ) {
	box->editbox.string.cursor = strlen ( box->editbox.string.buf );
}

void sf_editbox_set_num ( sf_editbox *box, long value ) {
	snprintf ( box->editbox.string.buf, box->editbox.string.len, "%ld", value );
	sf_editbox_set_cursor ( box );
}

void sf_combo_set ( sf_combo *combo, long value ) {
	int i;

	for (i = 0; i < combo->varn; ++i) {
		if (combo->vars[i].key == value) {
			combo->index = i;
			return;
		}
	}
}

/******************************************************************************/

/**
 * Draw method for 'Checkbox' widget
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
void sf_checkbox_draw ( struct sf_widget *this ) {
        sf_checkbox *box = SF_CHECKBOX ( this );

	if ( draw_disabled )
		return;

	draw_carret ( this );
	set_attr_selectable ( this );
	widget_mvwaddstr ( this, box->checked ? "[*] " : "[ ] " );
	widget_waddstr ( this, this->visible, this->text );

	/* Move to the place for the asterisk */
	wmove ( this->wnd, this->ori, this->col + 1 );
}

/**
 * Run method for 'Checkbox' widget
 *
 * @v this	The widget
 * @v key	Key pressed to process
 *
 * @ret		Status code
 */
int sf_checkbox_run ( struct sf_widget *this, int key) {
	sf_checkbox	*box = SF_CHECKBOX ( this );
	int	 	checked = -1;

	switch ( key ) {
	case CR:
	case LF:
	case ' ':
		checked = !box->checked;
		/* fall-thru */
	case '+':
		if ( checked == -1 )
			checked = 1;
		/* fall-thru */
	case '-':
		if ( checked == -1 )
			checked = 0;

		box->checked = checked;
		sf_widget_action ( this );
		sf_checkbox_draw ( this );
		return 0;

	default:
		return sf_widget_default_run ( key );
	}

}

/******************************************************************************/

/**
 * Draw method for 'Combo' widget
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
void sf_combo_draw ( struct sf_widget *this ) {
	sf_combo	*combo = SF_COMBO ( this );

	if ( draw_disabled )
		return;

	draw_carret ( this );
	set_attr_selectable ( this );
	widget_mvwaddstr ( this, this->text );

	widget_waddstr ( this, this->visible, combo->vars[combo->index].value );
}

/**
 * Run method for 'Combo' widget
 *
 * @v this	The widget
 * @v key	Key pressed to process
 *
 * @ret		Status code
 */
int sf_combo_run ( struct sf_widget *this, int key ) {
	sf_combo	*combo = SF_COMBO ( this );
	int		index = -1;
	int index_orig = combo->index;
	int rc;

	switch ( key ) {
	case CR:
	case LF:
	case ' ':
	case '+':
		index = combo->index + 1;
		/* fall-thru */
	case '-':
		if ( index == -1 )
			index = combo->index + combo->varn - 1;

		combo->index = index % combo->varn;
		rc = sf_widget_action ( this );
		if (rc & RC_FAIL) {
			/* If the action fails, reset to previous value */
			combo->index = index_orig;
		} else {
			sf_combo_draw ( this );
		}

		return rc;

	default:
		return sf_widget_default_run ( key );
	}
}

/******************************************************************************/

/**
 * Draw method for 'Static text' widget
 *
 * @v this	The widget
 *
 * @ret		Status code
 */
void sf_text_draw ( struct sf_widget *this ) {
	if ( draw_disabled )
		return;

	set_attr ( this->wnd, ATTR_TEXT, CPAIR_TEXT );
	widget_mvwaddstr ( this, this->text );
}

/******************************************************************************/

void sf_info_draw ( struct sf_widget *this ) {
	sf_info	*info = SF_INFO ( this );
	WINDOW *wnd = this->wnd;
	int	length;

	/* Info objects change frequently. If the contents have shrunk,
	 * then wipe out the old row */
	length = 0;
	if ( this->text != NULL )
		length += strlen ( this->text );
	if ( info->value != NULL )
		length += strlen ( info->value );

	wmove ( wnd, this->ori, this->col + length );
	set_attr ( wnd, ATTR_TEXT, CPAIR_TEXT );
	while ( length < info->last_length-- )
		waddch ( wnd, this->wnd->attrs | ' ' );
	info->last_length = length;

	/* Render the new values */
	if ( this->text != NULL ) {
		set_attr ( wnd, ATTR_TEXT, CPAIR_TEXT );
		mvwaddstr ( wnd, this->ori, this->col, this->text );
	}
	if ( info->value != NULL ) {
		set_attr ( wnd, ATTR_INFO, CPAIR_INFO );
		waddstr ( wnd, info->value );
	}
}


void sf_info_change ( sf_info *this, const char *name, const char *fmt, ... )
{
	free ( this->widget.text );
	this->widget.text = NULL;
	free ( this->value );
	this->value = NULL;

	if ( name != NULL )
		this->widget.text = strdup ( name );

	if ( fmt != NULL ) {
		va_list args;
		int rc;

		va_start ( args, fmt );
		rc = vasprintf ( &this->value, fmt, args );
		/* Truncate to 58 characters */
		if ( rc > 57 ) {
			this->value[54] = this->value[55] = this->value[56] = '.';
			this->value[57] = '\0';
		}
		va_end ( args );
	}
}

/******************************************************************************/

void sf_list_draw ( sf_widget *this ) {
	sf_list		*list = SF_LIST ( this );

	/* If the list is selected, then window_run() takes care of drawing */
	if ( !this->selected )
		sf_widget_draw ( list->child );
}

void sf_list_scroll ( sf_widget *this, int delta ) {
	sf_list *list = SF_LIST ( this );

	/* Move the entire window, don't scroll it's children */
	list->child->wnd->ori_y += delta;
}

int sf_list_run ( sf_widget *this, int key ) {
	sf_list *list = SF_LIST ( this );

	list->child->transient = TRUE;
	return sf_widget_run ( list->child, key );
}

/******************************************************************************/

static void sf_row_swap ( struct sf_widget *a, struct sf_widget *b )
{
	struct list_head *next = b->list.next;

	/* a-b-next:       move b after a, move a before next
	 * b-(next=a):     move b after a
	 * b-next-x-a:     move b after a, move a before next
	 * a-x-b-next:     move b after a, move a before next
	 */
	list_del ( &b->list );
	list_add ( &b->list, &a->list );

	if ( next != &a->list ) {
		list_del ( &a->list );
		list_add_tail ( &a->list, next );
	}
}

static int sf_row_cmp ( const struct sf_row *a, struct sf_row *b,
			 sf_row_cmp_t cmp )
{
	int ret;

	ret = ( a->widget.row - b->widget.row );
	if ( ret != 0 )
		return ret;

	return cmp ( a, b );
}

	
/* Row objects can be reordered on a form amongst other row objects, and
 * they load and store their setting value based on their position.
 *
 * It's intended that row objects are the only children of an sf_list object,
 * but you can do what you want of course.
 *
 * This function:
 *   - Normalises the row entries, so they're contiguous for all the visible
 *     sf_row objects in the window
 *   - Reorders the window linked list so that they're in increasing order,
 *     thereby allowing scrolling and reordering to progress.
 *
 * The intention is that once a user has added all of the sf_row entries to
 * a form, they should call sf_row_order() once
 */
void sf_row_order ( struct sf_widget *window, sf_row_cmp_t cmp )
{
	sf_widget *this, *prev;
	int row;

again:
	for ( prev = NULL, this = next_widget ( window );
	      this != window; this = next_widget ( this ) ) {
		if ( this->type != SFW_ROW || !this->visible )
			continue;

		if ( prev != NULL &&
		     sf_row_cmp ( SF_ROW ( this ),
				  SF_ROW ( prev ), cmp ) < 0 ) {
			/* Swap this and prev, then start again */
			sf_row_swap ( this, prev );
			goto again;
		}

		prev = this;
	}

	/* Fill out the row values, and dirty any elements that have moved */
	for ( row = 0, this = next_widget ( window ); this != window;
	      this = next_widget ( this ) ) {
		if ( this->type == SFW_ROW && this->visible ) {
			if ( row != this->row ) {
				int scroll = this->ori - this->row;

				this->row = row;
				this->ori = row + scroll;
			}
			++row;
		}
	}
}

int sf_row_run ( struct sf_widget *this, int key )
{
	sf_widget *w, *window;
	int row, ori;

	for ( w = prev_widget ( this ); w->type != SFW_WINDOW;
	      w = prev_widget ( w ) )
		;
	window = w;
	
	switch ( key ) {
	case '-':
		for ( w = prev_widget ( this ); w->type != SFW_WINDOW;
		      w = prev_widget ( w ) ) {
			if ( w->type == SFW_ROW && w->selectable )
				goto swap;
		}
		return 0;

	case '+':
		for ( w = next_widget ( this ); w->type != SFW_WINDOW;
		      w = next_widget ( w ) ) {
			if ( w->type == SFW_ROW && w->selectable )
				goto swap;
		}
		return 0;

	default:
		return sf_widget_default_run ( key );
	}

swap:
	sf_row_swap ( this, w );

	row = this->row;
	this->row = w->row;
	w->row = row;

	ori = this->ori;
	this->ori = w->ori;
	w->ori = ori;

	/* Move the viewport and redraw the widgets */
	if ( sf_window_viewport ( this, SF_WINDOW ( window ) ) )
		sf_widget_draw ( window );
	else {
		if ( w->visible )
			sf_widget_draw ( w );
		if ( this->visible )
			sf_widget_draw ( this );
	}

	return 0;
}

void sf_row_draw ( struct sf_widget *this )
{
	draw_carret ( this );

	set_attr ( this->wnd, ATTR_TEXT, CPAIR_TEXT );
	mvwprintw ( this->wnd, this->ori, this->col, "%d.  ", this->row + 1 );

	set_attr_selectable ( this );
	waddstr ( this->wnd, this->text );
}
