/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */
#ifndef _SFCONFIG_SF_WIDGETS_H
#define _SFCONFIG_SF_WIDGETS_H

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <config/general.h>
#include <curses.h>
#include <gpxe/list.h>
#include <gpxe/editbox.h>

/** @file
 *
 * Set of widgets for gPXE core settings operation, based on muCurses
 *
 */

/* A brief manual.
 *
 * To add a new widget:
 * 1. Take an sf_widget structure.
 * 2. Initialize it by sf_<widget type>_init function.
 * 3. Attach it to widgets list by sf_widget_attach(), if required.
 * 4. Show it by calling its "draw" method.
 *
 * The following widgets are provided:
 *
 * 1. Window. The 'next' field points to the list of widgets it contains.
 * 2. Button.
 * 3. Settings editbox - a widget for reading and editing gPXE core setting.
 * 4. Checkbox that can enable/disable other widgets and(or)
 *    check/uncheck other checkboxes.
 * 5. Choice list.
 * 6. Static text.
 */

/** Color pairs */
#define CPAIR_TEXT		1
#define CPAIR_DISABLED		2
#define CPAIR_CURRENT		3
#define CPAIR_INFO		4
#define ATTR_TEXT	        A_BOLD
#define ATTR_DISABLED		0
#define ATTR_CURRENT		A_BOLD
#define ATTR_INFO		A_BOLD

/** Return values */
#define RC_SUCCESS	0
#define RC_FAIL		0x80		/** generic failure */
#define RC_INVAL	(RC_FAIL+1)	/** Invalid parameters */
#define RC_GOUP		(RC_FAIL+2)	/** Go to previous widget */
#define RC_GODOWN	(RC_FAIL+3)	/** Go to next widget */
#define RC_GOLEFT	(RC_FAIL+4)	/** Go to left widget */
#define RC_GORIGHT	(RC_FAIL+5)	/** Go to right widget */
#define RC_TAB		(RC_FAIL+6)	/** Cycle through widgets */
#define RC_REVERT	(RC_FAIL+7)	/** Undo editing */
#define RC_COMPLETE	(RC_FAIL+8)	/** Widget activity is finished */
#define RC_QUIT		(RC_FAIL+9)	/** Unwind window stack and return RC_QUIT */

/** Widget type */
typedef enum sf_widget_type {
	SFW_WINDOW = 0,		/** Window */
	SFW_EDITBOX = 1,	/** Settings editbox */
	SFW_CHECKBOX = 2,	/** Checkbox */
	SFW_BUTTON = 3,		/** Button */
	SFW_COMBO = 4,		/** Combo box */
	SFW_TEXT = 5,		/** Static text */
	SFW_INFO = 6,		/** Static editbox */
	SFW_NEWLINE = 7,	/** Dummy widget "newline" */
	SFW_LIST = 8,		/** List widget */
	SFW_ROW = 9,		/** Row widget */
	SFW_NITEMS,
} sf_widget_type;

typedef int8_t sf_bool;

struct sf_widget;

/** Generic widget method operations */
struct sf_widget_operations {
	void	( * load ) ( struct sf_widget *this );
	int	( * save ) ( struct sf_widget *this, struct sf_widget **error_out );
	void	( * draw )( struct sf_widget *this );
	void	( * scroll ) ( struct sf_widget *this );
	int	( * run )( struct sf_widget *this, int key );
	void	( * delete )( struct sf_widget *this );
};

enum sf_carret_type {
	SF_CARRET_NONE,
	SF_CARRET_MARGIN,	/* A > character in the left margin */
	SF_CARRET_WINDOW,	/* > < characters either side of the widget */
};

/** Generic widget structure type */
typedef struct sf_widget {
	struct list_head                list;
	enum sf_widget_type		type;
	WINDOW				*wnd;
	/* Text/Title/Caption */
	char				*text;
	const char			*help;
	/* Screen position */
	int				col;
	int				row;    /* absolute */
	int				ori;    /* scrolled */
	/* Settings */
	struct settings			*settings;
	struct setting			*setting;
	/* Non-selectable widgets are skipped on arrow keys navigating */
	sf_bool				selectable;
	enum sf_carret_type		carret;
	sf_bool				selected;
	sf_bool				visible;
	sf_bool				transient;
	/* Action data */
	void				*aux_data;
	int				aux_data2;
	/* Methods */
	struct sf_widget_operations	*op;
	int				( * action )( struct sf_widget *this );
	void				( * enter )( struct sf_widget *this );
	void				( * exit )( struct sf_widget *this );
} sf_widget;

static inline __attribute__ (( always_inline )) sf_widget * next_widget ( sf_widget *w )

{
	return container_of ( w->list.next, struct sf_widget, list );
}

static inline __attribute__ (( always_inline )) sf_widget * prev_widget ( sf_widget *w )

{
	return container_of ( w->list.prev, struct sf_widget, list );
}

#define SF_WINDOW( _w ) container_of ( _w, struct sf_window, widget )
#define SF_BUTTON( _w ) container_of ( _w, struct sf_button, widget )
#define SF_EDITBOX( _w ) container_of ( _w, struct sf_editbox, widget )
#define SF_CHECKBOX( _w ) container_of ( _w, struct sf_checkbox, widget )
#define SF_COMBO( _w ) container_of ( _w, struct sf_combo, widget )
#define SF_TEXT( _w ) container_of ( _w, struct sf_text, widget )
#define SF_INFO( _w ) container_of ( _w, struct sf_info, widget )
#define SF_LIST( _w ) container_of ( _w, struct sf_list, widget )
#define SF_ROW( _w ) container_of ( _w, struct sf_row, widget )

/** 'Window' widget */
typedef struct sf_window {
	sf_widget		widget;
	struct list_head	list;
	int			draw_border;
	int			scrollable;
	int			cursor;
} sf_window;

/** 'Advanced editbox' widget, inspired by gPXE 'config' command. */
typedef struct sf_editbox {
	sf_widget		widget;
	struct edit_box		editbox;

	/* Shadow character, '\0' if no shadowing required */
	char			shadow;
} sf_editbox;

/** 'Checkbox' widget */
typedef struct sf_checkbox {
	sf_widget		widget;
	sf_bool	 		checked;
} sf_checkbox;

/** 'Button' widget */
typedef struct sf_button {
	sf_widget		widget;
} sf_button;

typedef struct sf_option {
	const char		*value;
	long			key;
} sf_option;

/** 'Combo' widget */
typedef struct sf_combo {
	sf_widget		widget;
	const sf_option		*vars;
	int			varn;
	int			index;
} sf_combo;

/** 'Static' widget */
typedef struct sf_text {
	sf_widget		widget;
} sf_text;

typedef struct sf_info {
	sf_widget		widget;
	char			*value;
	int			last_length;
} sf_info;

typedef struct sf_list {
	sf_widget		widget;
	sf_widget		*child;
} sf_list;

typedef struct sf_row {
	sf_widget		widget;
} sf_row;

/* Join two widget lists together */
extern void sf_widget_attach ( sf_widget *parent, sf_widget *child );

/* Free widgets list */
extern void sf_widgets_list_free ( sf_widget *list );

enum sf_dialog_type {
	SFW_DIALOG_MSG,
	SFW_DIALOG_OK,
	SFW_DIALOG_OKCANCEL,
	SFW_DIALOG_EDITCANCEL,
};

/* Show an information dialog */
extern int sf_dialog ( sf_widget **dialog, const char *title,
		       enum sf_dialog_type dialog_type, int width, int height,
		       const char **lines, unsigned int line_count );

/* Initialize sf_widget structure for 'window' widget */
extern sf_widget *sf_window_init ( const char *text, unsigned int row,
				   unsigned int column, unsigned int width,
				   unsigned int height );

/* Initialize sf_widget structure for 'button' widget */
extern sf_widget *sf_button_init ( const char *label, WINDOW *wnd,
				   unsigned int row, unsigned int column );

/* Initialize sf_widget structure for 'settings editbox' widget */
extern sf_widget *sf_editbox_init ( WINDOW *wnd, unsigned int row,
				    unsigned int column, size_t len,
				    const char *desc, struct settings *settings,
				    struct setting *setting,
				    unsigned int width );

static inline __attribute__((always_inline))
char *sf_editbox_get ( sf_editbox *box )
{
	return box->editbox.string.buf;
}

extern void sf_editbox_set ( sf_editbox *box, char *value );
extern void sf_editbox_set_num ( sf_editbox *box, long value );
extern void sf_editbox_set_cursor ( sf_editbox *box );

/* Initialize a sf_widget structure for a 'checkbox' widget */
extern sf_widget *sf_checkbox_init ( WINDOW *wnd, int row, int column,
				     const char *desc,
				     struct settings *settings,
				     struct setting *setting );

/* Initialize an sf_widget structure for a 'list' widget */
extern sf_widget *sf_combo_init ( WINDOW *wnd, int row, int column,
				  const char *desc,
				  struct settings *settings,
				  struct setting *setting,
				  const sf_option *vars, int varn );

extern void sf_combo_set ( sf_combo *combo, long value );

/* Initialize an sf_widget structure for a 'static text' widget */
extern sf_widget *sf_text_init ( WINDOW *wnd, int row, int column,
				 const char *text );

extern sf_widget *sf_info_init ( WINDOW *wnd, unsigned int row,
				 unsigned int column, const char *text,
				 const char* value );

extern void  __attribute__ (( format ( printf, 3, 4 ) ))
sf_info_change ( sf_info *info, const char *name, const char *fmt, ... );

extern sf_widget *sf_list_init ( WINDOW *wnd, unsigned int row,
				 unsigned int column, unsigned int width,
				 unsigned int height );

extern sf_widget *sf_row_init ( WINDOW *wnd, unsigned int column,
				const char *text,
				struct settings *settings,
				struct setting *setting );

typedef int ( *sf_row_cmp_t ) ( const struct sf_row *, const struct sf_row *);
extern void sf_row_order ( sf_widget *window, sf_row_cmp_t );

static inline __attribute__((always_inline)) void
sf_widget_load ( sf_widget *widget )
{
	widget->op->load ( widget );
}

static inline __attribute__((always_inline)) int
sf_widget_save ( sf_widget *widget, sf_widget **error_out )
{
	return widget->op->save ( widget, error_out );
}

static inline __attribute__((always_inline)) int
sf_widget_run ( sf_widget *widget, int key )
{
	return widget->op->run ( widget, key );
}

static inline __attribute__((always_inline)) void
sf_widget_draw ( sf_widget *widget )
{
	widget->op->draw ( widget );
}

extern int sf_widget_action ( sf_widget *widget );

#endif /* _SFCONFIG_SF_WIDGETS_H */
