/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Copyright (C) 2006 Michael Brown <mbrown@fensystems.co.uk>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include <byteswap.h>
#include <gpxe/vsprintf.h>
#include <gpxe/socket.h>
#include <gpxe/iobuf.h>
#include <gpxe/xfer.h>
#include <gpxe/open.h>
#include <gpxe/scsi.h>
#include <gpxe/process.h>
#include <gpxe/uaccess.h>
#include <gpxe/tcpip.h>
#include <gpxe/settings.h>
#include <gpxe/features.h>
#include <gpxe/iscsi.h>
/** OKTL code **/
#include <gpxe/netdevice.h>
#include <gpxe/sf_settings.h>
#include <gpxe/keys.h>
#include <console.h>

/** @file
 *
 * iSCSI protocol
 *
 */

FEATURE ( FEATURE_PROTOCOL, "iSCSI", DHCP_EB_FEATURE_ISCSI, 1 );

/** iSCSI initiator name (explicitly specified) */
static char *iscsi_explicit_initiator_iqn;

/** Default iSCSI initiator name (constructed from hostname) */
static char *iscsi_default_initiator_iqn;

/** iSCSI initiator username */
static char *iscsi_initiator_username;

/** iSCSI initiator password */
static char *iscsi_initiator_password;

/** iSCSI target username */
static char *iscsi_target_username;

/** iSCSI target password */
static char *iscsi_target_password;

static void iscsi_start_tx ( struct iscsi_session *iscsi );
static void iscsi_start_login ( struct iscsi_session *iscsi );
static void iscsi_start_data_out ( struct iscsi_session *iscsi,
				   unsigned int datasn );
static void iscsi_start_nop_out ( struct iscsi_session *iscsi );

/**
 * Finish receiving PDU data into buffer
 *
 * @v iscsi		iSCSI session
 */
static void iscsi_rx_buffered_data_done ( struct iscsi_session *iscsi ) {
	free ( iscsi->rx_buffer );
	iscsi->rx_buffer = NULL;
}

/**
 * Free iSCSI session
 *
 * @v refcnt		Reference counter
 */
static void iscsi_free ( struct refcnt *refcnt ) {
	struct iscsi_session *iscsi =
		container_of ( refcnt, struct iscsi_session, refcnt );

	free ( iscsi->cfg_target_address );
	free ( iscsi->new_target_address );
	free ( iscsi->target_iqn );
	free ( iscsi->initiator_username );
	free ( iscsi->initiator_password );
	free ( iscsi->target_username );
	free ( iscsi->target_password );
	chap_finish ( &iscsi->chap );
	iscsi_rx_buffered_data_done ( iscsi );
	free ( iscsi );
}

/**
 * Open iSCSI transport-layer connection
 *
 * @v iscsi		iSCSI session
 * @ret rc		Return status code
 */
static int iscsi_open_connection ( struct iscsi_session *iscsi ) {
	struct sockaddr_tcpip target;
	int rc;

	assert ( iscsi->conn_state == ISCSI_CONN_DEAD );
	assert ( iscsi->tx_state == ISCSI_TX_IDLE );
	assert ( iscsi->rx_state == ISCSI_RX_BHS );
	assert ( iscsi->rx_offset == 0 );

	/* Open socket */
	memset ( &target, 0, sizeof ( target ) );
	target.st_port = htons ( iscsi->target_port );
	if ( ( rc = xfer_open_named_socket ( &iscsi->socket, SOCK_STREAM,
					     ( struct sockaddr * ) &target,
					     iscsi->target_address,
					     NULL ) ) != 0 ) {
		DBGC ( iscsi, "iSCSI %p could not open socket: %s\n",
		       iscsi, strerror ( rc ) );
		return rc;
	}

	/* Enter security negotiation phase */
	iscsi->status = ( ISCSI_STATUS_SECURITY_NEGOTIATION_PHASE |
			  ISCSI_STATUS_STRINGS_SECURITY );

	/* Assign fresh initiator task tag */
	iscsi->itt++;

	/* Initiate login */
	iscsi->conn_state = ISCSI_CONN_LOGGING_IN;
	iscsi_start_login ( iscsi );

	return 0;
}

/**
 * Close iSCSI transport-layer connection
 *
 * @v iscsi		iSCSI session
 * @v rc		Reason for close
 *
 * Closes the transport-layer connection.  (The session state will then
 * be reset by iscsi_session_finish(), called from iscsi_socket_close()
 * via xfer_close().)
 */
static void iscsi_close_connection ( struct iscsi_session *iscsi, int rc ) {

	/* Close all data transfer interfaces */
	intf_restart ( &iscsi->socket, rc );
}

/**
 * Reset iSCSI session state
 *
 * @v iscsi		iSCSI session
 *
 * Closes the transport-layer connection and resets the session state
 * ready to attempt a fresh login.
 */
static void iscsi_session_finish ( struct iscsi_session *iscsi ) {

	/* Clear connection status */
	iscsi->conn_state = ISCSI_CONN_DEAD;
	iscsi->status = 0;

	/* Deauthenticate target */
	iscsi->target_auth_ok = 0;

	/* Reset TX and RX state machines */
	iscsi->tx_state = ISCSI_TX_IDLE;
	iscsi->rx_state = ISCSI_RX_BHS;
	iscsi->rx_offset = 0;

	/* Free any temporary dynamically allocated memory */
	chap_finish ( &iscsi->chap );
	iscsi_rx_buffered_data_done ( iscsi );
}

/**
 * Mark iSCSI SCSI operation as complete
 *
 * @v iscsi		iSCSI session
 * @v rc		Return status code
 *
 * Note that iscsi_scsi_done() will not close the connection, and must
 * therefore be called only when the internal state machines are in an
 * appropriate state, otherwise bad things may happen on the next call
 * to iscsi_issue().  The general rule is to call iscsi_scsi_done()
 * only at the end of receiving a PDU; at this point the TX and RX
 * engines should both be idle.
 */
static void iscsi_scsi_done ( struct iscsi_session *iscsi, int rc ) {

	assert ( iscsi->tx_state == ISCSI_TX_IDLE );

	iscsi->command = NULL;
	iscsi->rc = rc;
}

/**
 * Mark iSCSI logout operation as complete
 *
 * @v iscsi		iSCSI session
 * @v rc		Return status code
 *
 * See notes for iscsi_scsi_done().
 */
static void iscsi_logout_done ( struct iscsi_session *iscsi, int rc ) {

	assert ( iscsi->tx_state == ISCSI_TX_IDLE );

	/* Reset target address/port to the configured (or permanently
	 * redirected) settings -- if we were only temporarily redirected,
	 * we want to start the next login by connecting to the original
	 * server, not the one we were temporarily redirected to this time.
	 */
	iscsi->target_address = iscsi->cfg_target_address;
	iscsi->target_port = iscsi->cfg_target_port;

	iscsi->rc = rc;
}

/****************************************************************************
 *
 * iSCSI SCSI command issuing
 *
 */

/**
 * Build iSCSI SCSI command BHS
 *
 * @v iscsi		iSCSI session
 *
 * We don't currently support bidirectional commands (i.e. with both
 * Data-In and Data-Out segments); these would require providing code
 * to generate an AHS, and there doesn't seem to be any need for it at
 * the moment.
 */
static void iscsi_start_command ( struct iscsi_session *iscsi ) {
	struct iscsi_bhs_scsi_command *command = &iscsi->tx_bhs.scsi_command;

	assert ( ! ( iscsi->command->data_in && iscsi->command->data_out ) );

	/* Construct BHS and initiate transmission */
	iscsi_start_tx ( iscsi );
	command->opcode = ISCSI_OPCODE_SCSI_COMMAND;
	command->flags = ( ISCSI_FLAG_FINAL |
			   ISCSI_COMMAND_ATTR_SIMPLE );
	if ( iscsi->command->data_in )
		command->flags |= ISCSI_COMMAND_FLAG_READ;
	if ( iscsi->command->data_out )
		command->flags |= ISCSI_COMMAND_FLAG_WRITE;
	/* lengths left as zero */
	command->lun = iscsi->lun;
	command->itt = htonl ( ++iscsi->itt );
	command->exp_len = htonl ( iscsi->command->data_in_len |
				   iscsi->command->data_out_len );
	command->cmdsn = htonl ( iscsi->cmdsn );
	command->expstatsn = htonl ( iscsi->statsn + 1 );
	memcpy ( &command->cdb, &iscsi->command->cdb, sizeof ( command->cdb ));
	DBGC2 ( iscsi, "iSCSI %p start " SCSI_CDB_FORMAT " %s %#zx\n",
		iscsi, SCSI_CDB_DATA ( command->cdb ),
		( iscsi->command->data_in ? "in" : "out" ),
		( iscsi->command->data_in ?
		  iscsi->command->data_in_len :
		  iscsi->command->data_out_len ) );
}

/**
 * Receive data segment of an iSCSI SCSI response PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 */
static int iscsi_rx_scsi_response ( struct iscsi_session *iscsi,
				    const void *data, size_t len,
				    size_t remaining ) {
	struct iscsi_bhs_scsi_response *response
		= &iscsi->rx_bhs.scsi_response;
	int sense_offset;

	/* Capture the sense response code as it floats past, if present */
	sense_offset = ISCSI_SENSE_RESPONSE_CODE_OFFSET - iscsi->rx_offset;
	if ( ( sense_offset >= 0 ) && len ) {
		iscsi->command->sense_response =
			* ( ( char * ) data + sense_offset );
	}

	/* Wait for whole SCSI response to arrive */
	if ( remaining )
		return 0;
	
	/* Record SCSI status code */
	iscsi->command->status = response->status;

	/* Check for errors */
	if ( response->response != ISCSI_RESPONSE_COMMAND_COMPLETE )
		return -EIO;

	/* Mark as completed */
	iscsi_scsi_done ( iscsi, 0 );
	return 0;
}

/**
 * Receive data segment of an iSCSI data-in PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 */
static int iscsi_rx_data_in ( struct iscsi_session *iscsi,
			      const void *data, size_t len,
			      size_t remaining ) {
	struct iscsi_bhs_data_in *data_in = &iscsi->rx_bhs.data_in;
	unsigned long offset;

	/* Copy data to data-in buffer */
	offset = ntohl ( data_in->offset ) + iscsi->rx_offset;
	assert ( iscsi->command != NULL );
	assert ( iscsi->command->data_in );
	assert ( ( offset + len ) <= iscsi->command->data_in_len );
	copy_to_user ( iscsi->command->data_in, offset, data, len );

	/* Wait for whole SCSI response to arrive */
	if ( remaining )
		return 0;

	/* Mark as completed if status is present */
	if ( data_in->flags & ISCSI_DATA_FLAG_STATUS ) {
		assert ( ( offset + len ) == iscsi->command->data_in_len );
		assert ( data_in->flags & ISCSI_FLAG_FINAL );
		iscsi->command->status = data_in->status;
		/* iSCSI cannot return an error status via a data-in */
		iscsi_scsi_done ( iscsi, 0 );
	}

	return 0;
}

/**
 * Receive data segment of an iSCSI R2T PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 */
static int iscsi_rx_r2t ( struct iscsi_session *iscsi,
			  const void *data __unused, size_t len __unused,
			  size_t remaining __unused ) {
	struct iscsi_bhs_r2t *r2t = &iscsi->rx_bhs.r2t;

	/* Record transfer parameters and trigger first data-out */
	iscsi->ttt = ntohl ( r2t->ttt );
	iscsi->transfer_offset = ntohl ( r2t->offset );
	iscsi->transfer_len = ntohl ( r2t->len );
	iscsi_start_data_out ( iscsi, 0 );

	return 0;
}

/**
 * Build iSCSI data-out BHS
 *
 * @v iscsi		iSCSI session
 * @v datasn		Data sequence number within the transfer
 *
 */
static void iscsi_start_data_out ( struct iscsi_session *iscsi,
				   unsigned int datasn ) {
	struct iscsi_bhs_data_out *data_out = &iscsi->tx_bhs.data_out;
	unsigned long offset;
	unsigned long remaining;
	unsigned long len;

	/* We always send 512-byte Data-Out PDUs; this removes the
	 * need to worry about the target's MaxRecvDataSegmentLength.
	 */
	offset = datasn * 512;
	remaining = iscsi->transfer_len - offset;
	len = remaining;
	if ( len > 512 )
		len = 512;

	/* Construct BHS and initiate transmission */
	iscsi_start_tx ( iscsi );
	data_out->opcode = ISCSI_OPCODE_DATA_OUT;
	if ( len == remaining )
		data_out->flags = ( ISCSI_FLAG_FINAL );
	ISCSI_SET_LENGTHS ( data_out->lengths, 0, len );
	data_out->lun = iscsi->lun;
	data_out->itt = htonl ( iscsi->itt );
	data_out->ttt = htonl ( iscsi->ttt );
	data_out->expstatsn = htonl ( iscsi->statsn + 1 );
	data_out->datasn = htonl ( datasn );
	data_out->offset = htonl ( iscsi->transfer_offset + offset );
	DBGC ( iscsi, "iSCSI %p start data out DataSN %#x len %#lx\n",
	       iscsi, datasn, len );
}

/**
 * Complete iSCSI data-out PDU transmission
 *
 * @v iscsi		iSCSI session
 *
 */
static void iscsi_data_out_done ( struct iscsi_session *iscsi ) {
	struct iscsi_bhs_data_out *data_out = &iscsi->tx_bhs.data_out;

	/* If we haven't reached the end of the sequence, start
	 * sending the next data-out PDU.
	 */
	if ( ! ( data_out->flags & ISCSI_FLAG_FINAL ) )
		iscsi_start_data_out ( iscsi, ntohl ( data_out->datasn ) + 1 );
}

/**
 * Build iSCSI data-out data segment
 *
 * @v iscsi		iSCSI session
 * @v iobuf		iobuffer to put data into.
 */
static void iscsi_tx_build_data_out ( struct iscsi_session *iscsi,
				      struct io_buffer *iobuf )
{
	struct iscsi_bhs_data_out *data_out = &iscsi->tx_bhs.data_out;
	unsigned long offset;
	size_t len;

	offset = ntohl ( data_out->offset );
	len = ISCSI_DATA_LEN ( data_out->lengths );

	assert ( iscsi->command != NULL );
	assert ( iscsi->command->data_out );
	assert ( ( offset + len ) <= iscsi->command->data_out_len );

	copy_from_user ( iob_put ( iobuf, len ),
			 iscsi->command->data_out, offset, len );
}

/****************************************************************************
 *
 * iSCSI login
 *
 */

/**
 * Build iSCSI login request strings
 *
 * @v iscsi		iSCSI session
 *
 * These are the initial set of strings sent in the first login
 * request PDU.  We want the following settings:
 *
 *     HeaderDigest=None
 *     DataDigest=None
 *     MaxConnections is irrelevant; we make only one connection anyway
 *     InitialR2T=Yes [1]
 *     ImmediateData is irrelevant; we never send immediate data
 *     MaxRecvDataSegmentLength=8192 (default; we don't care) [3]
 *     MaxBurstLength=262144 (default; we don't care) [3]
 *     FirstBurstLength=262144 (default; we don't care)
 *     DefaultTime2Wait=0 [2]
 *     DefaultTime2Retain=0 [2]
 *     MaxOutstandingR2T=1
 *     DataPDUInOrder=Yes
 *     DataSequenceInOrder=Yes
 *     ErrorRecoveryLevel=0
 *
 * [1] InitialR2T has an OR resolution function, so the target may
 * force us to use it.  We therefore simplify our logic by always
 * using it.
 *
 * [2] These ensure that we can safely start a new task once we have
 * reconnected after a failure, without having to manually tidy up
 * after the old one.
 *
 * [3] We are quite happy to use the RFC-defined default values for
 * these parameters, but some targets (notably OpenSolaris)
 * incorrectly assume a default value of zero, so we explicitly
 * specify the default values.
 */
static int iscsi_build_login_request_strings ( struct iscsi_session *iscsi,
					       void *data, size_t len ) {
	unsigned int used = 0;
	unsigned int i;
	const char *auth_method;
	char *initiator_iqn;

	if ( iscsi->status & ISCSI_STATUS_STRINGS_SECURITY ) {
		/* Default to allowing no authentication */
		auth_method = "None";
		/* If we have a credential to supply, permit CHAP */
		if ( iscsi->initiator_username )
			auth_method = "CHAP,None";
		/* If we have a credential to check, force CHAP */
		if ( iscsi->target_username )
			auth_method = "CHAP";
		initiator_iqn = iscsi_initiator_iqn ( iscsi->netdev );
		used += ssnprintf ( data + used, len - used,
				    "InitiatorName=%s%c"
				    "TargetName=%s%c"
				    "SessionType=Normal%c"
				    "AuthMethod=%s%c",
				    initiator_iqn,
				    0, iscsi->target_iqn, 0, 0,
				    auth_method, 0 );
		free ( initiator_iqn );
	}

	if ( iscsi->status & ISCSI_STATUS_STRINGS_CHAP_ALGORITHM ) {
		used += ssnprintf ( data + used, len - used, "CHAP_A=5%c", 0 );
	}

	if ( ( iscsi->status & ISCSI_STATUS_STRINGS_CHAP_RESPONSE ) ) {
		assert ( iscsi->initiator_username != NULL );
		used += ssnprintf ( data + used, len - used,
				    "CHAP_N=%s%cCHAP_R=0x",
				    iscsi->initiator_username, 0 );
		for ( i = 0 ; i < iscsi->chap.response_len ; i++ ) {
			used += ssnprintf ( data + used, len - used, "%02x",
					    iscsi->chap.response[i] );
		}
		used += ssnprintf ( data + used, len - used, "%c", 0 );
	}

	if ( ( iscsi->status & ISCSI_STATUS_STRINGS_CHAP_CHALLENGE ) ) {
		used += ssnprintf ( data + used, len - used,
				    "CHAP_I=%d%cCHAP_C=0x",
				    iscsi->chap_challenge[0], 0 );
		for ( i = 1 ; i < sizeof ( iscsi->chap_challenge ) ; i++ ) {
			used += ssnprintf ( data + used, len - used, "%02x",
					    iscsi->chap_challenge[i] );
		}
		used += ssnprintf ( data + used, len - used, "%c", 0 );
	}

	if ( iscsi->status & ISCSI_STATUS_STRINGS_OPERATIONAL ) {
		used += ssnprintf ( data + used, len - used,
				    "HeaderDigest=None%c"
				    "DataDigest=None%c"
				    "InitialR2T=Yes%c"
				    "MaxRecvDataSegmentLength=8192%c"
				    "MaxBurstLength=262144%c"
				    "DefaultTime2Wait=0%c"
				    "DefaultTime2Retain=0%c"
				    "MaxOutstandingR2T=1%c"
				    "DataPDUInOrder=Yes%c"
				    "DataSequenceInOrder=Yes%c"
				    "ErrorRecoveryLevel=0%c",
				    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
	}

	return used;
}

/**
 * Build iSCSI login request BHS
 *
 * @v iscsi		iSCSI session
 */
static void iscsi_start_login ( struct iscsi_session *iscsi ) {
	struct iscsi_bhs_login_request *request = &iscsi->tx_bhs.login_request;
	int len;

	/* Construct BHS and initiate transmission */
	iscsi_start_tx ( iscsi );
	request->opcode = ( ISCSI_OPCODE_LOGIN_REQUEST |
			    ISCSI_FLAG_IMMEDIATE );
	request->flags = ( ( iscsi->status & ISCSI_STATUS_PHASE_MASK ) |
			   ISCSI_LOGIN_FLAG_TRANSITION );
	/* version_max and version_min left as zero */
	len = iscsi_build_login_request_strings ( iscsi, NULL, 0 );
	ISCSI_SET_LENGTHS ( request->lengths, 0, len );
	request->isid_iana_en = htonl ( ISCSI_ISID_IANA |
					IANA_EN_FEN_SYSTEMS );
	/* isid_iana_qual left as zero */
	request->tsih = htons ( iscsi->tsih );
	request->itt = htonl ( iscsi->itt );
	/* cid left as zero */
	request->cmdsn = htonl ( iscsi->cmdsn );
	request->expstatsn = htonl ( iscsi->statsn + 1 );
}

/**
 * Complete iSCSI login request PDU transmission
 *
 * @v iscsi		iSCSI session
 *
 */
static void iscsi_login_request_done ( struct iscsi_session *iscsi ) {

	/* Clear any "strings to send" flags */
	iscsi->status &= ~ISCSI_STATUS_STRINGS_MASK;

	/* Free any dynamically allocated storage used for login */
	chap_finish ( &iscsi->chap );
}

/**
 * Transmit data segment of an iSCSI login request PDU
 *
 * @v iscsi		iSCSI session
 * @v iobuf		iobuffer to put data into.
 *
 * For login requests, the data segment consists of the login strings.
 */
static void iscsi_tx_build_login_request ( struct iscsi_session *iscsi,
					   struct io_buffer *iobuf )
{
	struct iscsi_bhs_login_request *request = &iscsi->tx_bhs.login_request;
	size_t len;

	len = ISCSI_DATA_LEN ( request->lengths );
	iscsi_build_login_request_strings ( iscsi, iob_put ( iobuf, len ),
					    len );
}

/**
 * Handle iSCSI TargetAddress text value
 *
 * @v iscsi		iSCSI session
 * @v value		TargetAddress value
 * @ret rc		Return status code
 */
static int iscsi_handle_targetaddress_value ( struct iscsi_session *iscsi,
					      const char *value ) {
	char *separator;

	DBGC ( iscsi, "iSCSI %p will redirect to %s\n", iscsi, value );

	/* Replace target address */
	free ( iscsi->new_target_address );
	iscsi->new_target_address = strdup ( value );
	if ( ! iscsi->new_target_address )
		return -ENOMEM;

	/* Replace target port */
	iscsi->new_target_port = ISCSI_PORT;
	separator = strchr ( iscsi->new_target_address, ':' );
	if ( separator ) {
		*separator++ = '\0';
		if ( separator[0] != '\0' )
			iscsi->new_target_port = strtoul ( separator, NULL, 0 );
	}

	return 0;
}

/**
 * Handle iSCSI AuthMethod text value
 *
 * @v iscsi		iSCSI session
 * @v value		AuthMethod value
 * @ret rc		Return status code
 */
static int iscsi_handle_authmethod_value ( struct iscsi_session *iscsi,
					   const char *value ) {

	/* Mark target as authenticated if no authentication required */
	if ( ! iscsi->target_username )
		iscsi->target_auth_ok = 1;

	/* If server requests CHAP, send the CHAP_A string */
	if ( strcmp ( value, "CHAP" ) == 0 ) {
		DBGC ( iscsi, "iSCSI %p initiating CHAP authentication\n",
		       iscsi );
		iscsi->status |= ISCSI_STATUS_STRINGS_CHAP_ALGORITHM;
	}

	return 0;
}

/**
 * Handle a 4-character base64-encoded group
 *
 * @v encoded		Base64-encoded input string
 * @v out		Byte array for up to 3 output octets
 * @ret n		Number of decoded octets, or -1 for error
 */
static int iscsi_decode_base64_group ( const char *encoded, uint8_t *out ) {
	uint32_t word = 0;
	int consumed = 0;
	int bit_pos;
	char c;

	/* Handle end of string */
	if ( *encoded == 0 )
		return 0;

	/* Each input character decodes to 6 bits of output */
	for ( bit_pos = 18 ; bit_pos >= 0 ; bit_pos -= 6 ) {
		c = *encoded++;

		if ( ( c >= 'A' ) && ( c <= 'Z' ) )
			c -= 'A';
		else if ( ( c >= 'a' ) && ( c <= 'z' ) )
			c = ( c - 'a' + 26 );
		else if ( ( c >= '0' ) && ( c <= '9' ) )
			c = ( c - '0' + 52 );
		else if ( c == '+' )
			c = 62;
		else if ( c == '/' )
			c = 63;
		else if ( ( c == '=' ) && ( bit_pos < 12 ) )
			break;
		else
			return -1;

		word |= ( (uint32_t)c << bit_pos );
		consumed++;
	}

	/* Write output octets */
	out[0] = ( word >> 16 ) & 0xff;
	out[1] = ( word >> 8 ) & 0xff;
	out[2] = word & 0xff;

	/* Input of "====" -> invalid (already returned -1)
	 *          "x===" -> invalid (already returned -1)
	 *          "xx==" -> consumed==2; result is 1 valid output octet
	 *          "xxx=" -> consumed==3; result is 2 valid output octets
	 *          "xxxx" -> consumed==4; result is 3 valid output octets
	 */
	return consumed - 1;
}

/**
 * Handle iSCSI CHAP_A text value
 *
 * @v iscsi		iSCSI session
 * @v value		CHAP_A value
 * @ret rc		Return status code
 */
static int iscsi_handle_chap_a_value ( struct iscsi_session *iscsi,
				       const char *value ) {

	unsigned int algorithm;
	char *endp;

	/* We only ever offer "5" (i.e. MD5) as an algorithm, so if
	 * the server responds with anything else it is a protocol
	 * violation (but it is an integer value, not a string).
	 */
	algorithm = strtoul ( value, &endp, 0 );
	if ( ( *endp != '\0' ) || ( algorithm != 5 ) ) {
		DBGC ( iscsi, "iSCSI %p got invalid CHAP algorithm \"%s\"\n",
		       iscsi, value );
		return -EPROTO;
	}

	return 0;
}

/**
 * Handle iSCSI CHAP_I text value
 *
 * @v iscsi		iSCSI session
 * @v value		CHAP_I value
 * @ret rc		Return status code
 */
static int iscsi_handle_chap_i_value ( struct iscsi_session *iscsi,
				       const char *value ) {
	unsigned int identifier;
	char *endp;
	int rc;

	/* The CHAP identifier is an integer value */
	identifier = strtoul ( value, &endp, 0 );
	if ( *endp != '\0' ) {
		DBGC ( iscsi, "iSCSI %p saw invalid CHAP identifier \"%s\"\n",
		       iscsi, value );
		return -EPROTO;
	}

	/* Prepare for CHAP with MD5 */
	chap_finish ( &iscsi->chap );
	if ( ( rc = chap_init ( &iscsi->chap, &md5_algorithm ) ) != 0 ) {
		DBGC ( iscsi, "iSCSI %p could not initialise CHAP: %s\n",
		       iscsi, strerror ( rc ) );
		return rc;
	}

	/* Identifier and secret are the first two components of the
	 * challenge.
	 */
	chap_set_identifier ( &iscsi->chap, identifier );
	if ( iscsi->initiator_password ) {
		chap_update ( &iscsi->chap, iscsi->initiator_password,
			      strlen ( iscsi->initiator_password ) );
	}

	return 0;
}

/**
 * Handle iSCSI CHAP_C text value
 *
 * @v iscsi		iSCSI session
 * @v value		CHAP_C value
 * @ret rc		Return status code
 */
static int iscsi_handle_chap_c_value ( struct iscsi_session *iscsi,
				       const char *value ) {
	char buf[3];
	char *endp;
	uint8_t byte[3];
	unsigned int i;
	int n;

	/* The CHAP challenge is a large binary value, hex or base64 encoded */
	if ( ( value[0] == '0' ) && ( tolower ( value[1] ) == 'x' ) ) {
		/* Strip leading "0x" */
		value += 2;

		/* Process challenge an octet at a time */
		for ( ; ( value[0] && value[1] ) ; value += 2 ) {
			memcpy ( buf, value, 2 );
			buf[2] = 0;
			byte[0] = strtoul ( buf, &endp, 16 );
			if ( *endp != '\0' ) {
				DBGC ( iscsi, "iSCSI %p saw invalid CHAP "
				       "challenge byte \"%s\"\n", iscsi, buf );
				return -EPROTO;
			}
			chap_update ( &iscsi->chap, &byte[0], 1 );
		}
	} else if ( ( value[0] == '0' ) && ( tolower ( value[1] ) == 'b' ) ) {
		/* Strip leading "0b" */
		value += 2;

		/* Process challenge one base64-encoded group at a time */
		do {
			n = iscsi_decode_base64_group ( value, &byte[0] );
			if ( n < 0 ) {
				DBGC ( iscsi, "iSCSI %p saw invalid CHAP "
				       "base64 \"%s\"\n", iscsi, value );
				return -EPROTO;
			}
			else if ( n > 0 ) {
				chap_update ( &iscsi->chap, &byte[0], n );
			}
			value += 4;
		} while ( n == 3 );
	} else {
		DBGC ( iscsi, "iSCSI %p saw invalid CHAP challenge \"%s\"\n",
		       iscsi, value );
		return -EPROTO;
	}

	/* Build CHAP response */
	DBGC ( iscsi, "iSCSI %p sending CHAP response\n", iscsi );
	chap_respond ( &iscsi->chap );
	iscsi->status |= ISCSI_STATUS_STRINGS_CHAP_RESPONSE;

	/* Send CHAP challenge, if applicable */
	if ( iscsi->target_username ) {
		iscsi->status |= ISCSI_STATUS_STRINGS_CHAP_CHALLENGE;
		/* Generate CHAP challenge data */
		for ( i = 0 ; i < sizeof ( iscsi->chap_challenge ) ; i++ ) {
			iscsi->chap_challenge[i] = random();
		}
	}

	return 0;
}

/**
 * Handle iSCSI CHAP_N text value
 *
 * @v iscsi		iSCSI session
 * @v value		CHAP_N value
 * @ret rc		Return status code
 */
static int iscsi_handle_chap_n_value ( struct iscsi_session *iscsi,
				       const char *value ) {

	/* The target username isn't actually involved at any point in
	 * the authentication process; it merely serves to identify
	 * which password the target is using to generate the CHAP
	 * response.  We unnecessarily verify that the username is as
	 * expected, in order to provide mildly helpful diagnostics if
	 * the target is supplying the wrong username/password
	 * combination.
	 */
	if ( iscsi->target_username &&
	     ( strcmp ( iscsi->target_username, value ) != 0 ) ) {
		DBGC ( iscsi, "iSCSI %p target username \"%s\" incorrect "
		       "(wanted \"%s\")\n",
		       iscsi, value, iscsi->target_username );
		return -EACCES;
	}

	return 0;
}

/**
 * Handle iSCSI CHAP_R text value
 *
 * @v iscsi		iSCSI session
 * @v value		CHAP_R value
 * @ret rc		Return status code
 */
static int iscsi_handle_chap_r_value ( struct iscsi_session *iscsi,
				       const char *value ) {
	char buf[3];
	char *endp;
	uint8_t byte[3];
	unsigned int i;
	int j;
	int n;
	int rc;

	/* Generate CHAP response for verification */
	chap_finish ( &iscsi->chap );
	if ( ( rc = chap_init ( &iscsi->chap, &md5_algorithm ) ) != 0 ) {
		DBGC ( iscsi, "iSCSI %p could not initialise CHAP: %s\n",
		       iscsi, strerror ( rc ) );
		return rc;
	}
	chap_set_identifier ( &iscsi->chap, iscsi->chap_challenge[0] );
	if ( iscsi->target_password ) {
		chap_update ( &iscsi->chap, iscsi->target_password,
			      strlen ( iscsi->target_password ) );
	}
	chap_update ( &iscsi->chap, &iscsi->chap_challenge[1],
		      ( sizeof ( iscsi->chap_challenge ) - 1 ) );
	chap_respond ( &iscsi->chap );

	/* The CHAP response is a large binary value, hex or base64 encoded */
	if ( ( value[0] == '0' ) && ( tolower ( value[1] ) == 'x' ) ) {
		/* Strip leading "0x" */
		value += 2;

		/* Check CHAP response length */
		if ( strlen ( value ) != ( 2 * iscsi->chap.response_len ) ) {
			DBGC ( iscsi, "iSCSI %p invalid CHAP response "
			       "length\n", iscsi );
			return -EPROTO;
		}

		/* Process response an octet at a time */
		for ( i = 0 ; ( value[0] && value[1] ) ; value += 2, i++ ) {
			memcpy ( buf, value, 2 );
			buf[2] = 0;
			byte[0] = strtoul ( buf, &endp, 16 );
			if ( *endp != '\0' ) {
				DBGC ( iscsi, "iSCSI %p saw invalid CHAP "
				       "response byte \"%s\"\n", iscsi, buf );
				return -EPROTO;
			}
			if ( byte[0] != iscsi->chap.response[i] ) {
				DBGC ( iscsi, "iSCSI %p saw incorrect CHAP "
				       "response\n", iscsi );
				return -EACCES;
			}
		}
		assert ( i == iscsi->chap.response_len );
	} else if ( ( value[0] == '0' ) && ( tolower ( value[1] ) == 'b' ) ) {
		/* Strip leading "0b" */
		value += 2;

		/* Check CHAP response length */
		if ( strlen ( value ) !=
		     ( ( ( iscsi->chap.response_len + 2 ) / 3 ) * 4 ) ) {
			DBGC ( iscsi, "iSCSI %p invalid CHAP response "
			       "length\n", iscsi );
			return -EPROTO;
		}

		/* Process response one base64-encoded group at a time */
		i = 0;
		do {
			n = iscsi_decode_base64_group ( value, &byte[0] );
			if ( n < 0 ) {
				DBGC ( iscsi, "iSCSI %p saw invalid CHAP "
				       "response base64 \"%s\"\n", iscsi,
				       value );
				return -EPROTO;
			}
			for ( j = 0 ; j < n ; j++ ) {
				if ( byte[j] != iscsi->chap.response[i + j] ) {
					DBGC ( iscsi, "iSCSI %p saw incorrect "
					       "CHAP response\n", iscsi );
					return -EACCES;
				}
			}
			value += 4;
			i += n;
		} while ( n == 3 );
		assert ( i == iscsi->chap.response_len );
	} else {
		DBGC ( iscsi, "iSCSI %p saw invalid CHAP response \"%s\"\n",
		       iscsi, value );
		return -EPROTO;
	}

	/* Mark session as authenticated */
	iscsi->target_auth_ok = 1;

	return 0;
}

/** An iSCSI text string that we want to handle */
struct iscsi_string_type {
	/** String key
	 *
	 * This is the portion up to and including the "=" sign,
	 * e.g. "InitiatorName=", "CHAP_A=", etc.
	 */
	const char *key;
	/** Handle iSCSI string value
	 *
	 * @v iscsi		iSCSI session
	 * @v value		iSCSI string value
	 * @ret rc		Return status code
	 */
	int ( * handle ) ( struct iscsi_session *iscsi, const char *value );
};

/** iSCSI text strings that we want to handle */
static struct iscsi_string_type iscsi_string_types[] = {
	{ "TargetAddress=", iscsi_handle_targetaddress_value },
	{ "AuthMethod=", iscsi_handle_authmethod_value },
	{ "CHAP_A=", iscsi_handle_chap_a_value },
	{ "CHAP_I=", iscsi_handle_chap_i_value },
	{ "CHAP_C=", iscsi_handle_chap_c_value },
	{ "CHAP_N=", iscsi_handle_chap_n_value },
	{ "CHAP_R=", iscsi_handle_chap_r_value },
	{ NULL, NULL }
};

/**
 * Handle iSCSI string
 *
 * @v iscsi		iSCSI session
 * @v string		iSCSI string (in "key=value" format)
 * @ret rc		Return status code
 */
static int iscsi_handle_string ( struct iscsi_session *iscsi,
				 const char *string ) {
	struct iscsi_string_type *type;
	size_t key_len;
	int rc;

	for ( type = iscsi_string_types ; type->key ; type++ ) {
		key_len = strlen ( type->key );
		if ( strncmp ( string, type->key, key_len ) != 0 )
			continue;
		DBGC ( iscsi, "iSCSI %p handling %s\n", iscsi, string );
		if ( ( rc = type->handle ( iscsi,
					   ( string + key_len ) ) ) != 0 ) {
			DBGC ( iscsi, "iSCSI %p could not handle %s: %s\n",
			       iscsi, string, strerror ( rc ) );
			return rc;
		}
		return 0;
	}
	DBGC ( iscsi, "iSCSI %p ignoring %s\n", iscsi, string );
	return 0;
}

/**
 * Handle iSCSI strings
 *
 * @v iscsi		iSCSI session
 * @v string		iSCSI string buffer
 * @v len		Length of string buffer
 * @ret rc		Return status code
 */
static int iscsi_handle_strings ( struct iscsi_session *iscsi,
				  const char *strings, size_t len ) {
	size_t string_len;
	int rc;

	/* Handle each string in turn, taking care not to overrun the
	 * data buffer in case of badly-terminated data.
	 */
	while ( 1 ) {
		string_len = ( strnlen ( strings, len ) + 1 );
		if ( string_len > len )
			break;
		if ( ( rc = iscsi_handle_string ( iscsi, strings ) ) != 0 )
			return rc;
		strings += string_len;
		len -= string_len;
	}
	return 0;
}

/**
 * Receive PDU data into buffer
 *
 * @v iscsi		iSCSI session
 * @v data		Data to receive
 * @v len		Length of data
 * @ret rc		Return status code
 *
 * This can be used when the RX PDU type handler wishes to buffer up
 * all received data and process the PDU as a single unit.  The caller
 * is repsonsible for calling iscsi_rx_buffered_data_done() after
 * processing the data.
 */
static int iscsi_rx_buffered_data ( struct iscsi_session *iscsi,
				    const void *data, size_t len ) {

	/* Allocate buffer on first call */
	if ( ! iscsi->rx_buffer ) {
		iscsi->rx_buffer = malloc ( iscsi->rx_len );
		if ( ! iscsi->rx_buffer )
			return -ENOMEM;
	}

	/* Copy data to buffer */
	assert ( ( iscsi->rx_offset + len ) <= iscsi->rx_len );
	memcpy ( ( iscsi->rx_buffer + iscsi->rx_offset ), data, len );

	return 0;
}

/**
 * Convert iSCSI response status to return status code
 *
 * @v status_class	iSCSI status class
 * @v status_detail	iSCSI status detail
 * @ret rc		Return status code
 */
static int iscsi_status_to_rc ( unsigned int status_class,
				unsigned int status_detail ) {
	switch ( status_class ) {
	case ISCSI_STATUS_INITIATOR_ERROR :
		switch ( status_detail ) {
		case ISCSI_STATUS_INITIATOR_ERROR_AUTHENTICATION :
			return -EACCES;
		case ISCSI_STATUS_INITIATOR_ERROR_AUTHORISATION :
			return -EPERM;
		case ISCSI_STATUS_INITIATOR_ERROR_NOT_FOUND :
		case ISCSI_STATUS_INITIATOR_ERROR_REMOVED :
			return -ENODEV;
		default :
			return -ENOTSUP;
		}
	case ISCSI_STATUS_TARGET_ERROR :
		return -EIO;
	default :
		return -EINVAL;
	}
}

/**
 * Receive data segment of an iSCSI login response PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 */
static int iscsi_rx_login_response ( struct iscsi_session *iscsi,
				     const void *data, size_t len,
				     size_t remaining ) {
	struct iscsi_bhs_login_response *response
		= &iscsi->rx_bhs.login_response;
	int rc;

	/* Buffer up the PDU data */
	if ( ( rc = iscsi_rx_buffered_data ( iscsi, data, len ) ) != 0 ) {
		DBGC ( iscsi, "iSCSI %p could not buffer login response: %s\n",
		       iscsi, strerror ( rc ) );
		return rc;
	}
	if ( remaining )
		return 0;

	/* Process string data and discard string buffer */
	if ( ( rc = iscsi_handle_strings ( iscsi, iscsi->rx_buffer,
					   iscsi->rx_len ) ) != 0 )
		return rc;
	iscsi_rx_buffered_data_done ( iscsi );

	/* Check for login redirection */
	if ( response->status_class == ISCSI_STATUS_REDIRECT ) {
		printf ( "iSCSI target %s:%d redirects to %s:%d\n",
			 iscsi->target_address, iscsi->target_port,
			 iscsi->new_target_address, iscsi->new_target_port );
		switch ( response->status_detail ) {
		case ISCSI_STATUS_REDIRECT_TEMPORARY:
			iscsi->target_address = iscsi->new_target_address;
			iscsi->target_port = iscsi->new_target_port;
			break;
		case ISCSI_STATUS_REDIRECT_PERMANENT:
			free ( iscsi->cfg_target_address );
			iscsi->cfg_target_address =
				strdup ( iscsi->new_target_address );
			iscsi->cfg_target_port = iscsi->new_target_port;
			iscsi->target_address = iscsi->cfg_target_address;
			iscsi->target_port = iscsi->cfg_target_port;
			break;
		default:
			DBGC ( iscsi, "iSCSI %p got bad redirect status "
			       "detail %02x\n", iscsi,
			       response->status_detail );
			return -EPROTO;
		}
		iscsi->conn_state = ISCSI_CONN_REDIRECTING;
		iscsi_close_connection ( iscsi, 0 );
		return 0;
	}

	/* Check for fatal errors */
	if ( response->status_class != 0 ) {
		printf ( "iSCSI target %s:%d reports login error %02x%02x\n",
			 iscsi->target_address, iscsi->target_port,
			 response->status_class, response->status_detail );
		rc = iscsi_status_to_rc ( response->status_class,
					  response->status_detail );
		iscsi->instant_rc = rc;
		return rc;
	}

	/* Handle login transitions */
	if ( response->flags & ISCSI_LOGIN_FLAG_TRANSITION ) {
		switch ( response->flags & ISCSI_LOGIN_NSG_MASK ) {
		case ISCSI_LOGIN_NSG_OPERATIONAL_NEGOTIATION:
			iscsi->status =
				( ISCSI_STATUS_OPERATIONAL_NEGOTIATION_PHASE |
				  ISCSI_STATUS_STRINGS_OPERATIONAL );
			break;
		case ISCSI_LOGIN_NSG_FULL_FEATURE_PHASE:
			iscsi->status = ISCSI_STATUS_FULL_FEATURE_PHASE;
			break;
		default:
			DBGC ( iscsi, "iSCSI %p got invalid response flags "
			       "%02x\n", iscsi, response->flags );
			return -EPROTO;
		}
	}

	/* Send next login request PDU if we haven't reached the full
	 * feature phase yet.
	 */
	if ( ( iscsi->status & ISCSI_STATUS_PHASE_MASK ) !=
	     ISCSI_STATUS_FULL_FEATURE_PHASE ) {
		iscsi_start_login ( iscsi );
		return 0;
	}

	/* Check that target authentication was successful (if required) */
	if ( ! iscsi->target_auth_ok ) {
		DBGC ( iscsi, "iSCSI %p nefarious target tried to bypass "
		       "authentication\n", iscsi );
		return -EPROTO;
	}

	/* Connection is now fully active */
	iscsi->conn_state = ISCSI_CONN_ACTIVE;

	/* Reset retry and redirect counts */
	iscsi->retry_count = 0;
	iscsi->redirect_count = 0;

	/* Record TSIH for future reference */
	iscsi->tsih = ntohl ( response->tsih );
	
	/* Send the actual SCSI command */
	iscsi_start_command ( iscsi );

	return 0;
}

/****************************************************************************
 *
 * iSCSI NOP handling
 *
 */

/**
 * Receive data segment of an iSCSI NOP-In PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 */
static int iscsi_rx_nop_in ( struct iscsi_session *iscsi,
			     const void *data __unused, size_t len,
			     size_t remaining __unused ) {
	struct iscsi_bhs_nop_in *nop_in = &iscsi->rx_bhs.nop_in;

	/* Check ITT, TTT and DataSegmentLength */
	if ( nop_in->itt != ISCSI_RESERVED_TAG ) {
		DBGC ( iscsi, "iSCSI %p NOP-In with bad ITT\n", iscsi );
		return -EPROTO;
	}
	if ( nop_in->ttt == ISCSI_RESERVED_TAG ) {
		DBGC ( iscsi, "iSCSI %p NOP-In with bad TTT\n", iscsi );
		return -EPROTO;
	}
	if ( len != 0 ) {
		DBGC ( iscsi, "iSCSI %p NOP-In with non-empty data\n", iscsi );
		return -EPROTO;
	}

	/* Record transfer parameters and trigger NOP-Out */
	iscsi->ttt = ntohl ( nop_in->ttt );
	iscsi_start_nop_out ( iscsi );

	return 0;
}

/**
 * Build iSCSI NOP-Out BHS
 *
 * @v iscsi		iSCSI session
 *
 */
static void iscsi_start_nop_out ( struct iscsi_session *iscsi ) {
	struct iscsi_bhs_nop_out *nop_out = &iscsi->tx_bhs.nop_out;

	iscsi_start_tx ( iscsi );
	nop_out->opcode = ISCSI_OPCODE_NOP_OUT | ISCSI_FLAG_IMMEDIATE;
	nop_out->flags = ISCSI_FLAG_FINAL;
	nop_out->lun = iscsi->lun;
	nop_out->itt = ISCSI_RESERVED_TAG;
	nop_out->ttt = htonl ( iscsi->ttt );
	nop_out->cmdsn = htonl ( iscsi->cmdsn );
	nop_out->expstatsn = htonl ( iscsi->statsn + 1 );
	DBGC ( iscsi, "iSCSI %p start NOP out\n", iscsi );
}

/****************************************************************************
 *
 * iSCSI asynchronous message handling
 *
 */

/**
 * Receive data segment of an iSCSI asynchronous message PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 */
static int iscsi_rx_async_message ( struct iscsi_session *iscsi,
				    const void *data __unused,
				    size_t len __unused,
				    size_t remaining __unused ) {
	struct iscsi_bhs_async_message *async_msg =
						&iscsi->rx_bhs.async_message;

	switch ( async_msg->asyncevent ) {
	case ISCSI_ASYNC_EVENT_LOGOUT:
	case ISCSI_ASYNC_EVENT_NEGOTIATE:
		/* Target wants us to perform some action: either start a
		 * logout, or negotiate parameters (which we are allowed to
		 * handle by logging out instead).  Given our limited purpose,
		 * it should be sufficient to just ignore it (rather than
		 * adding more complexity to handle Logout Request / Response);
		 * the target should just give up and drop the connection
		 * anyway.
		 */
		DBGC ( iscsi, "iSCSI %p ignoring async message %d\n", iscsi,
		       async_msg->asyncevent );
		break;
	case ISCSI_ASYNC_EVENT_SCSI_AEN:
	case ISCSI_ASYNC_EVENT_DROP:
	case ISCSI_ASYNC_EVENT_DROP_ALL:
	case ISCSI_ASYNC_EVENT_VENDOR:
		/* Target is providing information: it can just be ignored. */
		break;
	default:
		DBGC ( iscsi, "iSCSI %p unknown async message %d\n", iscsi,
		       async_msg->asyncevent );
		return -EPROTO;
	}

	return 0;
}

/****************************************************************************
 *
 * iSCSI logout handling
 *
 */

/**
 * Build iSCSI Logout request BHS
 *
 * @v iscsi		iSCSI session
 *
 */
static void iscsi_start_logout_request ( struct iscsi_session *iscsi ) {
	struct iscsi_bhs_logout_request *request
		= &iscsi->tx_bhs.logout_request;

	iscsi_start_tx ( iscsi );
	request->opcode = ( ISCSI_OPCODE_LOGOUT_REQUEST |
			    ISCSI_FLAG_IMMEDIATE );
	request->flags = ( ISCSI_FLAG_FINAL |
			   ISCSI_LOGOUT_REASON_CLOSE_SESS );
	request->itt = htonl ( ++iscsi->itt );
	/* cid is filled in as 0 as we are asking to logout a session */
	request->cid = 0;
	request->cmdsn = htonl ( iscsi->cmdsn );
	request->expstatsn = htonl ( iscsi->statsn + 1 );
	DBGC ( iscsi, "iSCSI %p start logout request\n", iscsi );
}

/**
 * Receive data segment of an iSCSI logout response PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 */
static int iscsi_rx_logout_response ( struct iscsi_session *iscsi,
				      const void *data __unused,
				      size_t len __unused,
				      size_t remaining __unused ) {
	struct iscsi_bhs_logout_response *response
		= &iscsi->rx_bhs.logout_response;
	int resp_code;
	int rc;

	resp_code = response->response;
	DBGC ( iscsi, "iSCSI %p logout response: %d\n", iscsi, resp_code );
	switch ( resp_code ) {
	case ISCSI_LOGOUT_RESPONSE_SUCCESS:
		rc = 0;
		break;
	case ISCSI_LOGOUT_RESPONSE_CLEANUP_FAILED:
		rc = -EIO;
		break;
	default:
		rc = -EPROTO;
		break;
	}
	iscsi_logout_done ( iscsi, rc );
	iscsi_close_connection ( iscsi, rc );

	return 0;
}

/****************************************************************************
 *
 * iSCSI to socket interface
 *
 */

/**
 * Start up a new TX PDU
 *
 * @v iscsi		iSCSI session
 *
 * This initiates the process of sending a new PDU.  Only one PDU may
 * be in transit at any one time.
 */
static void iscsi_start_tx ( struct iscsi_session *iscsi ) {
	assert ( iscsi->tx_state == ISCSI_TX_IDLE );
	
	/* Initialise TX BHS */
	memset ( &iscsi->tx_bhs, 0, sizeof ( iscsi->tx_bhs ) );

	/* Flag TX engine to start transmitting */
	iscsi->tx_state = ISCSI_TX_BHS;
}

/**
 * Build basic header segment of an iSCSI PDU
 *
 * @v iscsi		iSCSI session
 * @v iobuf		iobuffer to put data into.
 */
static void iscsi_tx_build_bhs ( struct iscsi_session *iscsi,
				 struct io_buffer *iobuf )
{
	memcpy ( iob_put ( iobuf, sizeof ( iscsi->tx_bhs ) ),
		 &iscsi->tx_bhs, sizeof ( iscsi->tx_bhs ) );
}

/**
 * Build data segment of an iSCSI PDU
 *
 * @v iscsi		iSCSI session
 * @v iobuf		iobuffer to put data into.
 * 
 * Handle building a PDU data segment.  iscsi::tx_bhs
 * will be valid when this is called.
 */
static void iscsi_tx_build_data ( struct iscsi_session *iscsi,
				  struct io_buffer *iobuf )
{
	struct iscsi_bhs_common *common = &iscsi->tx_bhs.common;

	switch ( common->opcode & ISCSI_OPCODE_MASK ) {
	case ISCSI_OPCODE_DATA_OUT:
		iscsi_tx_build_data_out ( iscsi, iobuf );
		break;
	case ISCSI_OPCODE_LOGIN_REQUEST:
		iscsi_tx_build_login_request ( iscsi, iobuf );
		break;
	default:
		assert ( ISCSI_DATA_LEN ( common->lengths ) == 0 );
		break;
	}
}

/**
 * Transmit data padding of an iSCSI PDU
 *
 * @v iscsi		iSCSI session
 * @ret rc		Return status code
 * @v iobuf		iobuffer to put data into.
 * 
 * Handle building any data padding in a PDU data segment.
 * iscsi::tx_bhs will be valid when this is called.
 */
static void iscsi_tx_build_data_padding ( struct iscsi_session *iscsi,
					  struct io_buffer *iobuf )
{
 	struct iscsi_bhs_common *common = &iscsi->tx_bhs.common;
	size_t pad_len;
	
	pad_len = ISCSI_DATA_PAD_LEN ( common->lengths );
	if ( ! pad_len )
		return;

	memset ( iob_put ( iobuf, pad_len ), 0, pad_len );
}

/**
 * Complete iSCSI PDU transmission
 *
 * @v iscsi		iSCSI session
 *
 * Called when a PDU has been completely transmitted and the TX state
 * machine is about to enter the idle state.  iscsi::tx_bhs will be
 * valid for the just-completed PDU when this is called.
 */
static void iscsi_tx_done ( struct iscsi_session *iscsi ) {
	struct iscsi_bhs_common *common = &iscsi->tx_bhs.common;

	switch ( common->opcode & ISCSI_OPCODE_MASK ) {
	case ISCSI_OPCODE_DATA_OUT:
		iscsi_data_out_done ( iscsi );
		break;
	case ISCSI_OPCODE_LOGIN_REQUEST:
		iscsi_login_request_done ( iscsi );
		break;
	default:
		/* No action */
		break;
	}
}

/**
 * Transmit iSCSI PDU
 *
 * @v iscsi		iSCSI session
 * @v buf		Temporary data buffer
 * @v len		Length of temporary data buffer
 * 
 * Constructs data to be sent for the current TX state
 */
static void iscsi_tx_step ( struct process *process ) {
	struct iscsi_session *iscsi =
		container_of ( process, struct iscsi_session, process );
	struct iscsi_bhs_common *common = &iscsi->tx_bhs.common;
	struct io_buffer *iobuf;
	enum iscsi_tx_state next_state;
	size_t tx_len;
	int rc;

	/* All packets are sent in a single fragment */
	next_state = iscsi->tx_state;
	tx_len = 0;
	if ( next_state == ISCSI_TX_BHS ) {
		tx_len += sizeof ( iscsi->tx_bhs );
		next_state = ISCSI_TX_AHS;
	}
	if ( next_state == ISCSI_TX_AHS ) {
		tx_len += 0;
		next_state = ISCSI_TX_DATA;
	}
	if ( next_state == ISCSI_TX_DATA ) {
		tx_len += ISCSI_DATA_LEN ( common->lengths );
		next_state = ISCSI_TX_DATA_PADDING;
	}
	if ( next_state == ISCSI_TX_DATA_PADDING ) {
		tx_len += ISCSI_DATA_PAD_LEN ( common->lengths );
		next_state = ISCSI_TX_IDLE;
	}

	/* Check for window availability, if needed */
	if ( tx_len == 0 || xfer_window ( &iscsi->socket ) == 0 )
		return;

	iobuf = xfer_alloc_iob ( &iscsi->socket, tx_len );
	if ( !iobuf )
		return;
	
	next_state = iscsi->tx_state;
	if ( next_state == ISCSI_TX_BHS ) {
		iscsi_tx_build_bhs ( iscsi, iobuf );
		next_state = ISCSI_TX_AHS;
	}
	if ( next_state == ISCSI_TX_AHS ) {
		next_state = ISCSI_TX_DATA;
	}
	if ( next_state == ISCSI_TX_DATA ) {
		iscsi_tx_build_data ( iscsi, iobuf );
		next_state = ISCSI_TX_DATA_PADDING;
	}
	if ( next_state == ISCSI_TX_DATA_PADDING ) {
		iscsi_tx_build_data_padding ( iscsi, iobuf );
		next_state = ISCSI_TX_IDLE;
	}

	if ( ( rc = xfer_deliver_iob ( &iscsi->socket, iobuf ) ) != 0 ) {
		DBGC ( iscsi, "iSCSI %p could not transmit: %s\n",
		       iscsi, strerror ( rc ) );
		return;
	}

	/* Move to next state */
	iscsi->tx_state = next_state;
	if ( next_state == ISCSI_TX_IDLE )
		iscsi_tx_done ( iscsi );
}

/**
 * Receive basic header segment of an iSCSI PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 *
 * This fills in iscsi::rx_bhs with the data from the BHS portion of
 * the received PDU.
 */
static int iscsi_rx_bhs ( struct iscsi_session *iscsi, const void *data,
			  size_t len, size_t remaining __unused ) {
	memcpy ( &iscsi->rx_bhs.bytes[iscsi->rx_offset], data, len );
	if ( ( iscsi->rx_offset + len ) >= sizeof ( iscsi->rx_bhs ) ) {
		DBGC2 ( iscsi, "iSCSI %p received PDU opcode %#x len %#x\n",
			iscsi, iscsi->rx_bhs.common.opcode,
			ISCSI_DATA_LEN ( iscsi->rx_bhs.common.lengths ) );
	}
	return 0;
}

/**
 * Discard portion of an iSCSI PDU.
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 *
 * This discards data from a portion of a received PDU.
 */
static int iscsi_rx_discard ( struct iscsi_session *iscsi __unused,
			      const void *data __unused, size_t len __unused,
			      size_t remaining __unused ) {
	/* Do nothing */
	return 0;
}

/**
 * Receive data segment of an iSCSI PDU
 *
 * @v iscsi		iSCSI session
 * @v data		Received data
 * @v len		Length of received data
 * @v remaining		Data remaining after this data
 * @ret rc		Return status code
 *
 * Handle processing of part of a PDU data segment.  iscsi::rx_bhs
 * will be valid when this is called.
 */
static int iscsi_rx_data ( struct iscsi_session *iscsi, const void *data,
			   size_t len, size_t remaining ) {
	struct iscsi_bhs_common_response *response
		= &iscsi->rx_bhs.common_response;

	/* Update cmdsn and statsn */
	iscsi->cmdsn = ntohl ( response->expcmdsn );
	iscsi->statsn = ntohl ( response->statsn );

	switch ( response->opcode & ISCSI_OPCODE_MASK ) {
	case ISCSI_OPCODE_LOGIN_RESPONSE:
		return iscsi_rx_login_response ( iscsi, data, len, remaining );
	case ISCSI_OPCODE_SCSI_RESPONSE:
		return iscsi_rx_scsi_response ( iscsi, data, len, remaining );
	case ISCSI_OPCODE_DATA_IN:
		return iscsi_rx_data_in ( iscsi, data, len, remaining );
	case ISCSI_OPCODE_R2T:
		return iscsi_rx_r2t ( iscsi, data, len, remaining );
	case ISCSI_OPCODE_NOP_IN:
		return iscsi_rx_nop_in ( iscsi, data, len, remaining );
	case ISCSI_OPCODE_ASYNC_MESSAGE:
		return iscsi_rx_async_message ( iscsi, data, len, remaining );
	case ISCSI_OPCODE_LOGOUT_RESPONSE:
		return iscsi_rx_logout_response ( iscsi, data, len,
						  remaining );
	default:
		if ( remaining )
			return 0;
		DBGC ( iscsi, "iSCSI %p unknown opcode %02x\n", iscsi,
		       response->opcode );
		return -ENOTSUP;
	}
}

/**
 * Receive new data
 *
 * @v iscsi		iSCSI session
 * @v iobuf		I/O buffer
 * @v meta		Data transfer metadata
 * @ret rc		Return status code
 *
 * This handles received PDUs.  The receive strategy is to fill in
 * iscsi::rx_bhs with the contents of the BHS portion of the PDU,
 * throw away any AHS portion, and then process each part of the data
 * portion as it arrives.  The data processing routine therefore
 * always has a full copy of the BHS available, even for portions of
 * the data in different packets to the BHS.
 */
static int iscsi_socket_deliver ( struct iscsi_session *iscsi,
				  struct io_buffer *iobuf,
				  struct xfer_metadata *meta __unused ) {
	struct iscsi_bhs_common *common = &iscsi->rx_bhs.common;
	int ( * rx ) ( struct iscsi_session *iscsi, const void *data,
		       size_t len, size_t remaining );
	enum iscsi_rx_state next_state;
	size_t frag_len;
	size_t remaining;
	int rc;

	while ( 1 ) {
		switch ( iscsi->rx_state ) {
		case ISCSI_RX_BHS:
			rx = iscsi_rx_bhs;
			iscsi->rx_len = sizeof ( iscsi->rx_bhs );
			next_state = ISCSI_RX_AHS;			
			break;
		case ISCSI_RX_AHS:
			rx = iscsi_rx_discard;
			iscsi->rx_len = 4 * ISCSI_AHS_LEN ( common->lengths );
			next_state = ISCSI_RX_DATA;
			break;
		case ISCSI_RX_DATA:
			rx = iscsi_rx_data;
			iscsi->rx_len = ISCSI_DATA_LEN ( common->lengths );
			next_state = ISCSI_RX_DATA_PADDING;
			break;
		case ISCSI_RX_DATA_PADDING:
			rx = iscsi_rx_discard;
			iscsi->rx_len = ISCSI_DATA_PAD_LEN ( common->lengths );
			next_state = ISCSI_RX_BHS;
			break;
		default:
			assert ( 0 );
			rc = -EINVAL;
			goto done;
		}

		frag_len = iscsi->rx_len - iscsi->rx_offset;
		if ( frag_len > iob_len ( iobuf ) )
			frag_len = iob_len ( iobuf );
		remaining = iscsi->rx_len - iscsi->rx_offset - frag_len;
		if ( ( rc = rx ( iscsi, iobuf->data, frag_len,
				 remaining ) ) != 0 ) {
			DBGC ( iscsi, "iSCSI %p could not process received "
			       "data: %s\n", iscsi, strerror ( rc ) );
			iscsi_close_connection ( iscsi, rc );
			iscsi_scsi_done ( iscsi, rc );
			goto done;
		}

		iscsi->rx_offset += frag_len;
		iob_pull ( iobuf, frag_len );

		/* If all the data for this state has not yet been
		 * received, stay in this state for now.
		 */
		if ( iscsi->rx_offset != iscsi->rx_len ) {
			rc = 0;
			goto done;
		}

		iscsi->rx_state = next_state;
		iscsi->rx_offset = 0;
	}

	return 0;
done:
	/* Free I/O buffer */
	free_iob ( iobuf );
	return rc;
}

/**
 * Handle stream connection closure
 *
 * @v iscsi		iSCSI session
 * @v rc		Reason for close
 *
 */
static void iscsi_socket_close ( struct iscsi_session *iscsi, int rc ) {
	long max_retries = 0;
	enum iscsi_conn_state prev_conn_state = iscsi->conn_state;

	/** Beg OKTL code **/
	if ( iscsi->netdev && iscsi->settings )
		max_retries = sf_sett_get_intz ( iscsi->settings,
						 &lun_retries_setting );
	if ( max_retries < 0 )
		max_retries = 0;
	if ( max_retries > ISCSI_MAX_RETRIES )
		max_retries = ISCSI_MAX_RETRIES;
	/** End OKTL code **/

	/* Even a graceful close counts as an error for iSCSI */
	if ( ! rc )
		rc = -ECONNRESET;

	/* Reset session state (socket is already closed, that's why we're
	 * here!)
	 */
	iscsi_session_finish ( iscsi );

	/* Retry connection if within the retry limit, otherwise fail */
	if ( prev_conn_state == ISCSI_CONN_LOGGING_OUT ) {
		DBGC ( iscsi, "iSCSI %p not retrying connection\n", iscsi );
		iscsi->instant_rc = rc;
		iscsi_scsi_done ( iscsi, rc );
	} else if ( prev_conn_state == ISCSI_CONN_REDIRECTING ) {
		if ( ++iscsi->redirect_count <= ISCSI_MAX_REDIRECTS ) {
			DBGC ( iscsi, "iSCSI %p redirecting\n", iscsi );
			if ( ( rc = iscsi_open_connection ( iscsi ) ) != 0 ) {
				DBGC ( iscsi, "iSCSI %p could not redirect: "
				       "%s\n", iscsi, strerror ( rc ) );
				iscsi_scsi_done ( iscsi, rc );
			}
		} else {
			printf ( "Giving up: too many redirections\n" );
			rc = -EIO;
			iscsi->instant_rc = rc;
			iscsi_scsi_done ( iscsi, rc );
		}
	} else if ( ++iscsi->retry_count <= max_retries ) {
		DBGC ( iscsi, "iSCSI %p retrying connection (retry #%d)\n",
		       iscsi, iscsi->retry_count );
		iscsi->target_address = iscsi->cfg_target_address;
		iscsi->target_port = iscsi->cfg_target_port;
		if ( ( rc = iscsi_open_connection ( iscsi ) ) != 0 ) {
			DBGC ( iscsi, "iSCSI %p could not reconnect: %s\n",
			       iscsi, strerror ( rc ) );
			iscsi_scsi_done ( iscsi, rc );
		}
	} else {
		DBGC ( iscsi, "iSCSI %p retry count exceeded\n", iscsi );
		iscsi->instant_rc = rc;
		iscsi_scsi_done ( iscsi, rc );
	}
}

/**
 * Handle redirection event
 *
 * @v iscsi		iSCSI session
 * @v type		Location type
 * @v args		Remaining arguments depend upon location type
 * @ret rc		Return status code
 */
static int iscsi_vredirect ( struct iscsi_session *iscsi, int type,
			     va_list args ) {
	va_list tmp;
	struct sockaddr *peer;

	/* Intercept redirects to a LOCATION_SOCKET and record the IP
	 * address for the iBFT.  This is a bit of a hack, but avoids
	 * inventing an ioctl()-style call to retrieve the socket
	 * address from a data-xfer interface.
	 *
	 * Only do this if we're connecting to the configured address,
	 * (i.e. not a temporary redirect).
	 */
	if ( type == LOCATION_SOCKET &&
	     iscsi->target_address == iscsi->cfg_target_address ) {
		va_copy ( tmp, args );
		( void ) va_arg ( tmp, int ); /* Discard "semantics" */
		peer = va_arg ( tmp, struct sockaddr * );
		memcpy ( &iscsi->target_sockaddr, peer,
			 sizeof ( iscsi->target_sockaddr ) );
		va_end ( tmp );
	}

	return xfer_vreopen ( &iscsi->socket, type, args );
}

/** iSCSI socket interface operations */
static struct interface_operation iscsi_socket_operations[] = {
	INTF_OP ( xfer_deliver, struct iscsi_session *, iscsi_socket_deliver ),
	INTF_OP ( xfer_vredirect, struct iscsi_session *, iscsi_vredirect ),
	INTF_OP ( intf_close, struct iscsi_session *, iscsi_socket_close ),
};

/** iSCSI socket interface descriptor */
static struct interface_descriptor iscsi_socket_desc =
	INTF_DESC ( struct iscsi_session, socket, iscsi_socket_operations );

/****************************************************************************
 *
 * iSCSI command issuing
 *
 */

/**
 * Run until current command is complete
 *
 * @v iscsi		iSCSI session
 * @ret rc		Return status code
 */
static int iscsi_run ( struct iscsi_session *iscsi ) {
	int key;
	iscsi->rc = -EINPROGRESS;
	while ( iscsi->rc == -EINPROGRESS ) {
		step();
		if ( iskey() ) {
			key = getchar();
			if ( key == CTRL_C )
				iscsi->rc = -ECANCELED;
		}
	}
	return iscsi->rc;
}

/**
 * Issue SCSI command
 *
 * @v scsi		SCSI device
 * @v command		SCSI command
 * @ret rc		Return status code
 */
static int iscsi_command ( struct scsi_device *scsi,
			   struct scsi_command *command ) {
	struct iscsi_session *iscsi =
		container_of ( scsi->backend, struct iscsi_session, refcnt );
	int rc;

	/* Drain RX queue before processing command, to ensure that we e.g.
	 * respond to NOPs before starting anything new.
	 */
	netdev_rx_drain ( iscsi->netdev );

	/* Record SCSI command */
	iscsi->command = command;

	/* Connection may have been reset during a logout; that's OK */
	if ( ( iscsi->instant_rc == -ECONNRESET ) &&
	     ( ( iscsi->conn_state == ISCSI_CONN_LOGGING_OUT ) ||
	       ( iscsi->conn_state == ISCSI_CONN_DEAD ) ) )
		iscsi->instant_rc = 0;

	/* Abort immediately if we have a recorded permanent failure */
	if ( iscsi->instant_rc ) {
		rc = iscsi->instant_rc;
		goto done;
	}

	/* Issue command or open connection as appropriate */
	if ( iscsi->conn_state == ISCSI_CONN_ACTIVE ) {
		iscsi_start_command ( iscsi );
	} else {
		if ( ( rc = iscsi_open_connection ( iscsi ) ) != 0 )
			goto done;
	}

	rc = iscsi_run ( iscsi );

 done:
	iscsi->command = NULL;
	return rc;
}

/**
 * Issue logout
 *
 * @v scsi		SCSI device
 * @v command		SCSI command
 * @ret rc		Return status code
 */
int iscsi_logout ( struct scsi_device *scsi ) {
	struct iscsi_session *iscsi =
		container_of ( scsi->backend, struct iscsi_session, refcnt );
	int rc;

	/* Abort immediately if we have a recorded permanent failure */
	if ( iscsi->instant_rc ) {
		rc = iscsi->instant_rc;
		goto done;
	}

	/* Issue logout, or just ignore if not currently open */
	if ( iscsi->conn_state == ISCSI_CONN_ACTIVE ) {
		iscsi->conn_state = ISCSI_CONN_LOGGING_OUT;
		iscsi_start_logout_request ( iscsi );
		rc = iscsi_run ( iscsi );
	} else {
		rc = 0;
	}

 done:
	iscsi_logout_done ( iscsi, rc );
	iscsi_close_connection ( iscsi, 0 );
	return rc;
}

static int iscsi_detached_command ( struct scsi_device *scsi __unused,
				    struct scsi_command *command __unused ) {
	return -ENODEV;
}

/**
 * Shut down iSCSI interface
 *
 * @v scsi		SCSI device
 */
void iscsi_detach ( struct scsi_device *scsi ) {
	struct iscsi_session *iscsi =
		container_of ( scsi->backend, struct iscsi_session, refcnt );

	iscsi_close_connection ( iscsi, 0 );
	process_del ( &iscsi->process );
	scsi->command = iscsi_detached_command;
	ref_put ( scsi->backend );
	scsi->backend = NULL;
}

/****************************************************************************
 *
 * Instantiator
 *
 */

#if 0
/**
 * Set iSCSI authentication details
 *
 * @v iscsi		iSCSI session
 * @v initiator_username Initiator username, if any
 * @v initiator_password Initiator password, if any
 * @v target_username	Target username, if any
 * @v target_password	Target password, if any
 * @ret rc		Return status code
 */
static int iscsi_set_auth ( struct iscsi_session *iscsi,
			    const char *initiator_username,
			    const char *initiator_password,
			    const char *target_username,
			    const char *target_password ) {

	/* Check for initiator or target credentials */
	if ( initiator_username || initiator_password ||
	     target_username || target_password ) {

		/* We must have at least an initiator username+password */
		if ( ! ( initiator_username && initiator_password ) )
			goto invalid_auth;

		/* Store initiator credentials */
		iscsi->initiator_username = strdup ( initiator_username );
		if ( ! iscsi->initiator_username )
			return -ENOMEM;
		iscsi->initiator_password = strdup ( initiator_password );
		if ( ! iscsi->initiator_password )
			return -ENOMEM;

		/* Check for target credentials */
		if ( target_username || target_password ) {

			/* We must have target username+password */
			if ( ! ( target_username && target_password ) )
				goto invalid_auth;

			/* Store target credentials */
			iscsi->target_username = strdup ( target_username );
			if ( ! iscsi->target_username )
				return -ENOMEM;
			iscsi->target_password = strdup ( target_password );
			if ( ! iscsi->target_password )
				return -ENOMEM;
		}
	}

	return 0;

 invalid_auth:
	DBGC ( iscsi, "iSCSI %p invalid credentials: initiator "
	       "%sname,%spw, target %sname,%spw\n", iscsi,
	       ( initiator_username ? "" : "no " ),
	       ( initiator_password ? "" : "no " ),
	       ( target_username ? "" : "no " ),
	       ( target_password ? "" : "no " ) );
	return -EINVAL;
}
#endif

/**
 * Attach iSCSI interface
 *
 * @v scsi		SCSI device
 * @v root_path		iSCSI root path (as per RFC4173)
 * @ret rc		Return status code
 */
int iscsi_attach ( struct scsi_device *scsi, const char *root_path ) {
	struct iscsi_session	*iscsi;
	uint64_t		lun;
	int			rc;

	/* Allocate and initialise structure */
	iscsi = zalloc ( sizeof ( *iscsi ) );
	if ( ! iscsi )
		return -ENOMEM;
	ref_init ( &iscsi->refcnt, iscsi_free );
	intf_init ( &iscsi->socket, &iscsi_socket_desc, &iscsi->refcnt );
	process_init ( &iscsi->process, iscsi_tx_step, &iscsi->refcnt );

	/* Parse root path */
	if ( ( rc = iscsi_parse_root_path ( root_path,
					    &iscsi->cfg_target_address,
					    &iscsi->cfg_target_port,
					    &lun,
					    &iscsi->target_iqn ) ) != 0 )
		goto err;
	/* iscsi->lun is a 64-bit host-endian representation of the SCSI SAM2
	 * LUN format.; a single level LUN of 0-255 (which is all we support)
	 * is 00,nn,00,00,00,00,00,00 */
	iscsi->lun = ((lun & 0xff) << 8);

	/** Beg OKTL code **/
	if ( ( iscsi->netdev = last_opened_netdev ( ) ) == NULL ) {
		DBGC ( iscsi, "iscsi_attach failed: no net device");
		rc = -ENODEV;
		goto err;
	}
	iscsi->settings = netdev_nvo_settings ( iscsi->netdev );
	if ( iscsi->settings == NULL ) {
		DBGC ( iscsi, "iscsi_attach aborted: net device %s has no "
		       "nvo settings", iscsi->netdev->name );
		rc = -ENOENT;
		goto err;
	}
	if ( sf_sett_get_intz ( iscsi->settings, &chap_setting ) ) {
		if ( sf_sett_fetcha_string ( iscsi->settings,
					     &username_setting,
					     &iscsi->initiator_username ) < 0 )
			iscsi->initiator_username = strdup("");

		if ( sf_sett_fetcha_string ( iscsi->settings,
					     &password_setting,
					     &iscsi->initiator_password ) < 0 )
			iscsi->initiator_password = strdup("");
	}

	if ( sf_sett_get_intz ( iscsi->settings, &chap_setting ) &&
	     sf_sett_get_intz ( iscsi->settings, &chap_target_setting ) ) {
		if ( sf_sett_fetcha_string ( iscsi->settings,
					     &reverse_username_setting,
					     &iscsi->target_username ) < 0 )
			iscsi->target_username = strdup("");
		
		if ( sf_sett_fetcha_string ( iscsi->settings,
					     &reverse_password_setting,
					     &iscsi->target_password ) < 0 )
			iscsi->target_password = strdup("");
	}

	/* Sanity checks */
	if ( ! iscsi->cfg_target_address ) {
		DBGC ( iscsi, "iSCSI %p does not yet support discovery\n",
		       iscsi );
		rc = -ENOTSUP;
		goto err;
	}
	if ( ! iscsi->target_iqn ) {
		DBGC ( iscsi, "iSCSI %p no target address supplied in %s\n",
		       iscsi, root_path );
		rc = -EINVAL;
		goto err;
	}

	/* Use configured target address/port for first login */
	iscsi->target_address = iscsi->cfg_target_address;
	iscsi->target_port = iscsi->cfg_target_port;

	/** Beg OKTL code: Print out connection information */
	char *initiator_iqn, *hostname;
	printf ( "iSCSI target name          : %s\n", iscsi->target_iqn );
	printf ( "iSCSI target address       : %s\n", iscsi->target_address );
	printf ( "iSCSI LUN                  : %d\n", (int)lun );
	printf ( "iSCSI port                 : %d\n", iscsi->target_port );
	initiator_iqn = iscsi_initiator_iqn ( iscsi->netdev );
	printf ( "iSCSI initiator name       : %s\n", initiator_iqn );
	free ( initiator_iqn );
	hostname = iscsi_initiator_hostname ( iscsi->netdev );
	printf ( "iSCSI initiator hostname   : %s\n", hostname );
	free ( hostname );
	
	DBGP ( "iSCSI credentials: initiator: <%s> <%s>\n",
	       iscsi->initiator_username,
	       iscsi->initiator_password );
	DBGP ( "iSCSI credentials: target: <%s> <%s>\n",
	       iscsi->target_username,
	       iscsi->target_password );

	printf ( "\n" );
	/** End OKTL code **/

	/* Attach parent interface, mortalise self, and return */
	scsi->backend = ref_get ( &iscsi->refcnt );
	scsi->command = iscsi_command;
	scsi->lun = iscsi->lun;
	rc = 0;
err:
	ref_put ( &iscsi->refcnt );
	return rc;
}

/****************************************************************************
 *
 * Settings
 *
 */

/** An iSCSI string setting */
struct iscsi_string_setting {
	/** Setting */
	struct setting *setting;
	/** String to update */
	char **string;
	/** String prefix */
	const char *prefix;
};


/** iSCSI string settings */
static struct iscsi_string_setting iscsi_string_settings[] = {
	{
		.setting = &initiator_iqn_setting,
		.string = &iscsi_explicit_initiator_iqn,
		.prefix = "",
	},
	{
		.setting = &username_setting,
		.string = &iscsi_initiator_username,
		.prefix = "",
	},
	{
		.setting = &password_setting,
		.string = &iscsi_initiator_password,
		.prefix = "",
	},
	{
		.setting = &reverse_username_setting,
		.string = &iscsi_target_username,
		.prefix = "",
	},
	{
		.setting = &reverse_password_setting,
		.string = &iscsi_target_password,
		.prefix = "",
	},
	{
		.setting = &hostname_setting,
		.string = &iscsi_default_initiator_iqn,
		.prefix = "iqn.2000-01.org.etherboot:",
	},
};

/**
 * Apply iSCSI setting
 *
 * @v setting		iSCSI string setting
 * @ret rc		Return status code
 */
static int apply_iscsi_string_setting ( struct iscsi_string_setting *setting ){
	size_t prefix_len;
	int setting_len;
	size_t len;
	int check_len;
	char *p;

	/* Free old string */
	free ( *setting->string );
	*setting->string = NULL;

	/* Allocate new string */
	prefix_len = strlen ( setting->prefix );
	setting_len = fetch_setting_len ( NULL, setting->setting );
	if ( setting_len < 0 ) {
		/* Missing settings are not errors; leave strings as NULL */
		return 0;
	}
	len = ( prefix_len + setting_len + 1 );
	p = *setting->string = malloc ( len );
	if ( ! p )
		return -ENOMEM;

	/* Fill new string */
	strcpy ( p, setting->prefix );
	check_len = fetch_string_setting ( NULL, setting->setting,
					   ( p + prefix_len ),
					   ( len - prefix_len ) );
	assert ( check_len == setting_len );

	return 0;
}

/**
 * Apply iSCSI settings
 *
 * @ret rc		Return status code
 */
static int apply_iscsi_settings ( void ) {
	struct iscsi_string_setting *setting;
	unsigned int i;
	int rc;

	for ( i = 0 ; i < ( sizeof ( iscsi_string_settings ) /
			    sizeof ( iscsi_string_settings[0] ) ) ; i++ ) {
		setting = &iscsi_string_settings[i];
		if ( ( rc = apply_iscsi_string_setting ( setting ) ) != 0 ) {
			DBG ( "iSCSI could not apply setting %s\n",
			      setting->setting->name );
			return rc;
		}
	}

	return 0;
}

/** iSCSI settings applicator */
struct settings_applicator iscsi_settings_applicator __settings_applicator = {
	.apply = apply_iscsi_settings,
};



/**
 * Get iSCSI initiator IQN
 *
 * @v netdev		netdevice
 * @ret rc		Return status code
 */
char * iscsi_initiator_iqn ( struct net_device *netdev ) {
	struct settings *settings_nvo, *settings_dhcp, *settings;
	char *iqn;
	char *buf = NULL;

	if ( netdev == NULL ) {
		DBG ( "Device to get Initiator IQN is not specified\n" );
		goto failed;
	}

	settings_nvo = netdev_nvo_settings ( netdev );
	settings_dhcp = netdev_dhcp_settings ( netdev );

	/* Explicit IQN */
	settings = ( sf_sett_get_intz ( settings_nvo,
					&initiator_iqn_dhcp_setting )
		     ? settings_dhcp : settings_nvo );
	if ( settings != NULL ) {
		if ( sf_sett_fetcha_string ( settings,
					     &initiator_iqn_setting,
					     &iqn ) > 0 )
			goto out;
	}

	/* Construct from dhcp hostname */
	if ( settings_dhcp != NULL ) 
		sf_sett_fetcha_string ( settings_dhcp,
					&hostname_setting,
					&buf );

failed:
	asprintf ( &iqn, "iqn.2000-09.org.etherboot:%s",
		   buf ? buf : "UNKNOWN" );
out:
	free ( buf );
	return iqn;
}

/**
 * Get iSCSI hostname
 *
 * Microsoft iSCSI initiator really, really wants a hostname in the
 * iBFT, which means we should try a little harder when option12
 * isn't provided in dhcp
 *
 * @v netdev		netdevice
 * @ret rc		Return status code
 */
char * iscsi_initiator_hostname ( struct net_device *netdev ) {
	struct settings *settings_dhcp;
	char *hostname, *p, *initiator_iqn = NULL;

	if ( netdev == NULL ) {
		DBG ( "Device to get Initiator IQN is not specified\n" );
		goto failed;
	}

	/* If the user provided a hostname over dhcp, use it */
	settings_dhcp = netdev_dhcp_settings ( netdev );
	if ( settings_dhcp != NULL ) {
		if ( sf_sett_fetcha_string ( settings_dhcp,
					     &hostname_setting,
					     &hostname ) > 0 )
			goto out;
	}

	/* Parse the IQN */
	initiator_iqn = iscsi_initiator_iqn ( netdev );
	if ( initiator_iqn != NULL &&
	     ( p = strstr ( initiator_iqn, ":" ) ) != NULL ) {
		hostname = strdup ( p + 1 );
		goto out;
	}

failed:
	/* The iBFT really, really wants a hostname */
	hostname = strdup ( "UNKNOWN" );

out:
	if ( initiator_iqn != NULL )
		free ( initiator_iqn );
	return hostname;
}
       
	
	
